	.include "m32u4def.inc"

#define clock 8.0

#define VCC 3.3
#define LOW_BATT 0.95
#define LOW_BATT_CNT 3

.equ RAM_ADDR_BITS=16	;16 for 23x512, 17 for 23x1024

#define ep_cs PORTE,6
#define ep_dc PORTD,7
#define ep_rs PORTD,4
#define ep_bs PIND,0

#define SPI_SS PORTB,5
#define SPI_SS_DDR DDRB,5

#define	CS		PORTB,4	/* MMC_CS */
#define CS_DDR	DDRB,4
#define SD_PWR	PORTC,6
#define SD_PWR_DDR DDRC,6
#define SD_CD	PIND,1
#define SD_CD_PULLUP PORTD,1

#define OPTSENSOR PINF,5
#define OPTSENSOR_MUX 5
#define OPTSENSOR_PULLUP PORTF,5
#define OPTSENSOR_DIGIN ADC5D
#define OPTSENSOR_DIGREG DIDR0
#define OPTSENSOR_THR 10			;ambient light threshold, defer processing if below

#define BATTSENSOR PINF,4
#define BATTSENSOR_MUX 4
#define BATTSENSOR_PULLUP PORTF,4
#define BATTSENSOR_DIGIN ADC4D
#define BATTSENSOR_DIGREG DIDR0

#define NOISE_RCVR PINF,6
#define NOISE_RCVR_MUX 6
#define NOISE_RCVR_PULLUP PORTF,6
#define NOISE_RCVR_DIGIN ADC6D
#define NOISE_RCVR_DIGREG DIDR0

#define status SREG
#define sp_low SPL
#define sp_high SPH

; EPD4IN2 commands
#define PANEL_SETTING                               0x00
#define POWER_SETTING                               0x01
#define POWER_OFF                                   0x02
#define POWER_OFF_SEQUENCE_SETTING                  0x03
#define POWER_ON                                    0x04
#define POWER_ON_MEASURE                            0x05
#define BOOSTER_SOFT_START                          0x06
#define DEEP_SLEEP                                  0x07
#define DATA_START_TRANSMISSION_1                   0x10
#define DATA_STOP                                   0x11
#define DISPLAY_REFRESH                             0x12
#define DATA_START_TRANSMISSION_2                   0x13
#define LUT_FOR_VCOM                                0x20 
#define LUT_WHITE_TO_WHITE                          0x21
#define LUT_BLACK_TO_WHITE                          0x22
#define LUT_WHITE_TO_BLACK                          0x23
#define LUT_BLACK_TO_BLACK                          0x24
#define PLL_CONTROL                                 0x30
#define TEMPERATURE_SENSOR_COMMAND                  0x40
#define TEMPERATURE_SENSOR_SELECTION                0x41
#define TEMPERATURE_SENSOR_WRITE                    0x42
#define TEMPERATURE_SENSOR_READ                     0x43
#define VCOM_AND_DATA_INTERVAL_SETTING              0x50
#define LOW_POWER_DETECTION                         0x51
#define TCON_SETTING                                0x60
#define RESOLUTION_SETTING                          0x61
#define GSST_SETTING                                0x65
#define GET_STATUS                                  0x71
#define AUTO_MEASUREMENT_VCOM                       0x80
#define READ_VCOM_VALUE                             0x81
#define VCM_DC_SETTING                              0x82
#define PARTIAL_WINDOW                              0x90
#define PARTIAL_IN                                  0x91
#define PARTIAL_OUT                                 0x92
#define PROGRAM_MODE                                0xA0
#define ACTIVE_PROGRAMMING                          0xA1
#define READ_OTP                                    0xA2
#define POWER_SAVING                                0xE3

#define WT_IDLE -1

#define pname_in_ram GPIOR0,0
#define cache_flag	 GPIOR0,1
#define valid_sect   GPIOR0,2
#define rnd_pics	 GPIOR0,3		;0-sequential, 1-random picture show
#define rd_src		 GPIOR0,4		;0-read from flash, 1-read from SPI RAM
#define card_fail	 GPIOR0,5		;1-too many file access errors on card
#define is_card		 GPIOR0,6		;0-no card inserted, 1-card inserted
#define ep_rdy		 GPIOR0,7

;following both parameters can be controlled by CONTROL.TXT file on memory card
#define default_show_dly 120		;delay in seconds until showing next picture (multiples of 8 and >=8)
;#define default_random_mode		;use random mode for showing pictures; uncomment for sequential mode 

;image size is hard coded
.equ x_=400
.equ y_=300

.equ img_plane_size=x_/2*y_
.equ ht_size=512			;size of decoding tree memory
.equ file_retry=5			;retry count for opening picture file 
.equ pic_retry=10			;retry count for picture decoding

.equ recent_entries=5		;number of entries to check repeated file names

.equ rd_buf_size=1024		;multiples of 512

#define cache_sector
#define legacy_support

/* MMC/SD command (SPI mode) */
#define CMD0	0		/* GO_IDLE_STATE */
#define CMD1	1		/* SEND_OP_COND */
#define	ACMD41	0x80+41	/* SEND_OP_COND (SDC) */
#define CMD8	8		/* SEND_IF_COND */
#define CMD9	9		/* SEND_CSD */
#define CMD10	10		/* SEND_CID */
#define CMD12	12		/* STOP_TRANSMISSION */
#define CMD13	13		/* SEND_STATUS */
#define ACMD13	0x80+13	/* SD_STATUS (SDC) */
#define CMD16	16		/* SET_BLOCKLEN */
#define CMD17	17		/* READ_SINGLE_BLOCK */
#define CMD18	18		/* READ_MULTIPLE_BLOCK */
#define CMD23	23		/* SET_BLOCK_COUNT */
#define	ACMD23	0x80+23	/* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24	24		/* WRITE_BLOCK */
#define CMD25	25		/* WRITE_MULTIPLE_BLOCK */
#define CMD32	32		/* ERASE_ER_BLK_START */
#define CMD33	33		/* ERASE_ER_BLK_END */
#define CMD38	38		/* ERASE */
#define CMD55	55		/* APP_CMD */
#define CMD58	58		/* READ_OCR */


/* Disk Status Bits (DSTATUS) */
#define STA_NOINIT		1	/* Drive not initialized */
#define STA_NODISK		2	/* No medium in the drive */
#define STA_PROTECT		4	/* Write protected */
; Disk Status Bit Positions
#define STA_NOINIT_bp	0	/* Drive not initialized */
#define STA_NODISK_bp	1	/* No medium in the drive */
#define STA_PROTECT_bp	2	/* Write protected */

/* MMC card type flags (MMC_GET_TYPE) */
#define CT_MMC		1		/* MMC ver 3 */
#define CT_SD1		2		/* SD ver 1 */
#define CT_SD2		4		/* SD ver 2 */
#define CT_SDC		(CT_SD1|CT_SD2)	/* SD */
#define CT_BLOCK	8		/* Block addressing */

/* Results of Disk Functions (DRESULT) */
#define	RES_OK		0		/* 0: Successful */
#define	RES_ERROR	1		/* 1: R/W Error */
#define	RES_WRPRT	2		/* 2: Write Protected */
#define	RES_NOTRDY	3		/* 3: Not Ready */
#define	RES_PARERR	4		/* 4: Invalid Parameter */

#define CTRL_SYNC			0	/* Complete pending write process (needed at _FS_READONLY == 0) */
#define GET_SECTOR_COUNT	1	/* Get media size (needed at _USE_MKFS == 1) */
#define GET_SECTOR_SIZE		2	/* Get sector size (needed at _MAX_SS != _MIN_SS) */
#define GET_BLOCK_SIZE		3	/* Get erase block size (needed at _USE_MKFS == 1) */

#define FS_FAT12	1
#define FS_FAT16	2
#define FS_FAT32	3

/* File attribute bits for directory entry */

#define	AM_RDO	0x01	/* Read only */
#define	AM_HID	0x02	/* Hidden */
#define	AM_SYS	0x04	/* System */
#define	AM_VOL	0x08	/* Volume label */
#define AM_LFN	0x0F	/* LFN entry */
#define AM_DIR	0x10	/* Directory */
#define AM_ARC	0x20	/* Archive */
#define AM_MASK	0x3F	/* Mask of defined bits */
#define AM_RDO_bp	0

#define	FA_OPENED	0x01
#define	FA_WPRT		0x02
#define	FA__WIP		0x40

#define _FS_FAT12
#define _FS_FAT16
#define _FS_FAT32

#ifdef _FS_FAT32
#define clust_sz 4
#else
#define clust_sz 2
#endif

#define BS_jmpBoot			0
#define BS_OEMName			3
#define BPB_BytsPerSec		11
#define BPB_SecPerClus		13
#define BPB_RsvdSecCnt		14
#define BPB_NumFATs			16
#define BPB_RootEntCnt		17
#define BPB_TotSec16		19
#define BPB_Media			21
#define BPB_FATSz16			22
#define BPB_SecPerTrk		24
#define BPB_NumHeads		26
#define BPB_HiddSec			28
#define BPB_TotSec32		32
#define BS_55AA				510

#define BS_DrvNum			36
#define BS_BootSig			38
#define BS_VolID			39
#define BS_VolLab			43
#define BS_FilSysType		54

#define BPB_FATSz32			36
#define BPB_ExtFlags		40
#define BPB_FSVer			42
#define BPB_RootClus		44
#define BPB_FSInfo			48
#define BPB_BkBootSec		50
#define BS_DrvNum32			64
#define BS_BootSig32		66
#define BS_VolID32			67
#define BS_VolLab32			71
#define BS_FilSysType32		82

#define MBR_Table			446

#define	DIR_Name			0
#define	DIR_Attr			11
#define	DIR_NTres			12
#define	DIR_CrtTime			14
#define	DIR_CrtDate			16
#define	DIR_FstClusHI		20
#define	DIR_WrtTime			22
#define	DIR_WrtDate			24
#define	DIR_FstClusLO		26
#define	DIR_FileSize		28

#define	FR_OK 				0
#define	FR_DISK_ERR			1
#define	FR_NOT_READY		2
#define	FR_NO_FILE			3
#define	FR_NOT_OPENED		4
#define	FR_NOT_ENABLED		5
#define	FR_NO_FILESYSTEM	6
#define FR_END_OF_FILE		7
#define FR_WRONG_PARAM		8
#define FR_WR_PROTECTED		9
#define FR_EXISTS_ALREADY	10

	.org 0
	rjmp start

	.org INT0addr
	sbi ep_rdy
	reti

	.org INT1addr
	reti

	.org WDTaddr
	push r16
	in r16,SREG
	push r16
	lds r16,(wd_cnt)
	inc r16
	sts (wd_cnt),r16
	pop r16
	out SREG,r16
	pop r16
	reti

start:
	ldi r16,LOW(RAMEND)
	out (SPL),r16
	ldi r16,HIGH(RAMEND)
	out (SPH),r16

	ldi r16,-1
	out PORTB,r16
	out PORTC,r16
	out PORTD,r16
	out PORTE,r16
	out PORTF,r16
	cbi PORTE,2

	sbi PORTB,0
	sbi CS
	ldi r16,1<<PB4|1<<PB2|1<<PB1|1<<PB0
	out DDRB,r16
	sbi SPI_SS
	sbi SPI_SS_DDR

	sbi SD_PWR_DDR

	sbi ep_cs
	sbi ep_dc
	sbi ep_rs
	sbi PORTD,5
	ldi r16,1<<PD7|1<<PD5|1<<PD4
	out DDRD,r16
	ldi r16,1<<6
	out DDRE,r16
	cbi SD_CD_PULLUP			;disable pull up on SD_CD

	cbi OPTSENSOR_PULLUP		;disable pull up on optical sensor input
	lds r16,(OPTSENSOR_DIGREG)
	ori r16,1<<OPTSENSOR_DIGIN	;disable digital input on optical sensor
	sts (OPTSENSOR_DIGREG),r16

	cbi BATTSENSOR_PULLUP		;disable pull up on battery sensor input
	lds r16,(BATTSENSOR_DIGREG)
	ori r16,1<<BATTSENSOR_DIGIN	;disable digital input on battery sensor
	sts (BATTSENSOR_DIGREG),r16

	cbi NOISE_RCVR_PULLUP
	lds r16,(NOISE_RCVR_DIGREG)
	ori r16,1<<NOISE_RCVR_DIGIN
	sts (NOISE_RCVR_DIGREG),r16

	ldi r16,1<<ISC10|1<<ISC01|1<<ISC00
	sts (EICRA),r16
	ldi r16,1<<INT1|1<<INT0
	out (EIMSK),r16

	ldi r16,1<<SPE|1<<MSTR
	out SPCR,r16
	ldi r16,1<<SPI2X
	out SPSR,r16

	ldi ZL,LOW(random)
	ldi ZH,HIGH(random)
	ldi r16,NOISE_RCVR_MUX|1<<REFS0
	sts (ADMUX),r16
	ldi r19,4
seed_lp2:
	ldi r18,8
seed_lp1:
	ldi r16,1<<ADEN|1<<ADSC|1<<ADIF|1<<ADPS2|1<<ADPS1
	sts (ADCSRA),r16
adc_get_bit:
	lds r16,(ADCSRA)
	sbrs r16,ADIF
	rjmp adc_get_bit
	lds r16,(ADCL)
	lds r0,(ADCH)
	lsr r16
	rol r17
	dec r18
	brne seed_lp1
	st Z+,r17
	dec r19
	brne seed_lp2
	ldi r16,$aa
	eor r17,r16
	breq seed1
	st -Z,r17
seed1:
	clr r16							;disable ADC
	sts (ADCSRA),r16
	sbi NOISE_RCVR_PULLUP
	lds r16,(NOISE_RCVR_DIGREG)
	andi r16,~(1<<NOISE_RCVR_DIGIN)
	sts (NOISE_RCVR_DIGREG),r16

	ldi r16,1<<SM1|1<<SE
	out SMCR,r16

 	ldi r16,1<<ACD
	out ACSR,r16
	ldi r16,1<<PRTWI|1<<PRTIM0|1<<PRTIM1|1<<PRADC
	sts (PRR0),r16
	ldi r16,1<<PRUSB|1<<PRTIM3|1<<PRUSART1
	sts (PRR1),r16

	wdr
	ldi r16,1<<WDCE|1<<WDE
	sts (WDTCSR),r16
	ldi r16,1<<WDIE|1<<WDP3|1<<WDP0		;enable watchdog timer for interrupt every 8 seconds
	sts (WDTCSR),r16
	sei

	sbi is_card
	cbi card_fail

	clr r16
	sts (lowbatt_cnt),r16

;do a single screen refresh with default settings
	cbi ep_rs
	rcall delay2ms
	sbi ep_rs
	rcall delay2ms
	ldi ZL,LOW(ep_init_std*2)
	ldi ZH,HIGH(ep_init_std*2)
	rcall ep_hndl_cmdstr
	ldi r16,DATA_START_TRANSMISSION_1
	rcall ep_clr_ram
	ldi r16,DATA_START_TRANSMISSION_2
	rcall ep_clr_ram
	rcall ep_disp_frame
	rcall delay2ms

restart:
	sbic SD_CD
	rjmp no_card 
	rcall check_optsensor
	brcs prepare_sleep1_j
	rcall read_config
	tst r16
	brne no_card
	rcall reset_pfile_ptr
	rjmp card_ins
start1:
	sbic SD_CD
	rjmp no_card
	rcall check_optsensor
prepare_sleep1_j:
	brcs prepare_sleep1
	call mount
	tst r16
	breq card_ins
no_card:
	sbi CS
	sbi SD_PWR
	sbis is_card
	rjmp prepare_sleep
	cbi is_card
	cbi rd_src				;read from flash
	ldi r16,LOW(insert_card_pic*2)
	sts (rd_ptr),r16
	ldi r16,HIGH(insert_card_pic*2)
	sts (rd_ptr+1),r16
	rcall show_pic
	tst r16
	breq restart
	rjmp prepare_sleep
card_ins:
	ldi r18,pic_retry		;retry count in case there's something wrong with the picture file
open_pic_file1:
	ldi r17,file_retry		;retry count in case there's something wrong with a path name
open_pic_file:
	push r18
	push r17
	rcall get_pic_fname
	pop r17
	pop r18
	tst r16
	breq show_picture
	dec r17
	brne open_pic_file
	sbi card_fail
	rjmp no_card
show_picture:
	push r18
	rcall show_pic_s
	pop r18
	tst r16
	brpl show_picture1
	dec r18
	brne open_pic_file1
	sbi CS
	sbi SD_PWR
	sbi card_fail
	rjmp show_pic_err1
show_picture1:
	breq prepare_sleep
	cpi r16,2
	breq no_card
show_pic_err:
	cbi is_card
show_pic_err1:
	cbi rd_src				;read from flash
	ldi r16,LOW(insert_card_pic*2)
	sts (rd_ptr),r16
	ldi r16,HIGH(insert_card_pic*2)
	sts (rd_ptr+1),r16
	rcall show_pic1

prepare_sleep:
	lds r18,(w_time)
	sbis is_card
prepare_sleep1:
	ldi r18,1

	ldi r16,BATTSENSOR_MUX
	rcall adc_
	subi r16,LOW(INT(1023*LOW_BATT/VCC))
	sbci r17,HIGH(INT(1023*LOW_BATT/VCC))
	ldi r16,0
	brcc do_sleep
	lds r16,(lowbatt_cnt)
	inc r16
	cpi r16,LOW_BATT_CNT
	brcc low_battery
do_sleep:
	sts (lowbatt_cnt),r16
	sbis PORTB,3
	rjmp do_sleep1
	rcall ep_disp_power_off
	ldi r16,1<<INT1
	out (EIMSK),r16
	rcall ep_disp_sleep
do_sleep1:
	cbi PORTB,3				;disable PullUp for MISO 
	wdr
	clr r16
	sts (wd_cnt),r16
sleep_lp:
	sleep

	sbis card_fail
	rjmp no_card_fail
	sbis SD_CD
	rjmp prepare_sleep1
	cbi card_fail
	cbi is_card
	rjmp prepare_sleep1

no_card_fail:
	clr r16
	sbic is_card
	inc r16
	sbis SD_CD
	dec r16
	tst r16
	brne continue

	lds r16,(wd_cnt)
	cp r16,r18
	brcs sleep_lp

continue:
	ldi r16,1<<INT1|1<<INT0
	out (EIMSK),r16
	sbi PORTB,3				;enable PullUp for MISO 
	sbis is_card
	rjmp restart
	rjmp start1

low_battery:
	cbi rd_src				;read from flash
	ldi r16,LOW(lowbatt_pic*2)
	sts (rd_ptr),r16
	ldi r16,HIGH(lowbatt_pic*2)
	sts (rd_ptr+1),r16
	rcall show_pic
	rcall ep_disp_power_off
	ldi r16,1<<INT1
	out (EIMSK),r16
	rcall ep_disp_sleep
	cbi PORTB,3				;disable PullUp for MISO
	wdr
	ldi r16,1<<WDCE|1<<WDE
	sts (WDTCSR),r16
	ldi r16,0
	sts (WDTCSR),r16
sleep_forever:
	sleep
	rjmp sleep_forever

show_pic_s:
	cpi r18,pic_retry
	brne show_pic1
show_pic:
	cbi ep_rs
	rcall delay2ms
	sbi ep_rs
	rcall delay2ms
	ldi ZL,LOW(ep_lut_clr*2)
	ldi ZH,HIGH(ep_lut_clr*2)
	rcall ep_hndl_cmdstr
show_pic1:
	rcall dec_jpg
	tst r16
	brne show_picx
	sbi SD_PWR
	sbis rd_src
	rjmp show_pic2
	sbis SD_CD
	rjmp show_pic2
	ldi r16,1
	rjmp show_picx
show_pic2:
	rcall ep_wt_idle
	ldi ZL,LOW(ep_init_volt_str*2)
	ldi ZH,HIGH(ep_init_volt_str*2)
	rcall ep_hndl_cmdstr
	ldi r18,1
	rcall plane1
	ldi ZL,LOW(ep_lut_grayxs_str*2)
	ldi ZH,HIGH(ep_lut_grayxs_str*2)
	rcall ep_hndl_cmdstr
	rcall ep_disp_frame
	ldi r18,4
	rcall plane1
	ldi ZL,LOW(ep_lut_grayxs1_str*2)
	ldi ZH,HIGH(ep_lut_grayxs1_str*2)
	rcall ep_hndl_cmdstr
	rcall ep_disp_frame
	ldi r16,0
	sbic SD_CD
	ldi r16,2
show_picx:
	ret

ep_disp_power_off:
	ldi ZL,LOW(ep_power_off_str*2)
	ldi ZH,HIGH(ep_power_off_str*2)
	rjmp ep_hndl_cmdstr

ep_disp_sleep:
	ldi ZL,LOW(ep_sleep_str*2)
	ldi ZH,HIGH(ep_sleep_str*2)
	rjmp ep_hndl_cmdstr

get_pic_fname_rst:
	rcall reset_pfile_ptr
get_pic_fname:
	sbi is_card
	sbi rd_src				;read from disk
	ldi ZL,LOW(pfile*2)
	ldi ZH,HIGH(pfile*2)
	cbi pname_in_ram
	call f_open
	tst r16
	brne get_pic_fname_x

	sbis rnd_pics
	rcall get_fname_seq
	sbic rnd_pics
	rcall get_fname_rnd

	ldi XL,LOW(buffer)
	ldi XH,HIGH(buffer)
get_pic_fname_cp:
	mov r16,r17
	or r16,r18
	breq get_pic_fname_rst
	ld r16,Z+
	cpi r16,' '
	brcs get_pic_fname_open_img
	st X+,r16
	subi r17,1
	sbci r18,0
	rjmp get_pic_fname_cp
get_pic_fname_open_img:
	ld r16,Z
	cpi r16,' '
	brcc get_pic_fname_open_img1
	adiw Z,1
	rjmp get_pic_fname_open_img
get_pic_fname_open_img1:
	clr r16
	st X+,r16
	subi ZL,LOW(sect_buf)
	sbci ZH,HIGH(sect_buf)
	ldi YL,LOW(pict_file_ptr)
	ldi YH,HIGH(pict_file_ptr)
	ld r16,Y
	add r16,ZL
	st Y+,r16
	ld r16,Y
	adc r16,ZH
	st Y+,r16
	ldi ZL,0
	ld r16,Y
	adc r16,ZL
	st Y+,r16
	ld r16,Y
	adc r16,ZL
	st Y+,r16

	ldi ZL,LOW(buffer)
	ldi ZH,HIGH(buffer)
	sbi pname_in_ram
	call f_open
	sbi CS
	tst r16
	brne get_pic_fname_x
	ldi r16,HIGH(sect_buf+rd_buf_size)
	sts (rd_ptr+1),r16
	clr r16
get_pic_fname_x:
	ret

reset_pfile_ptr:
	clr r16
	ldi YL,LOW(pict_file_ptr)
	ldi YH,HIGH(pict_file_ptr)
	std Y+3,r16
	std Y+2,r16
	std Y+1,r16
	st Y,r16	
	ret

cfile: .db "/CONFIG.TXT",0
pfile: .db "/PICTURES.TXT",0

read_config:
	ldi r17,10
read_conf_lp:
	push r17
	call mount
	pop r17
	tst r16
	breq read_config1
	dec r17
	brne read_conf_lp
	rjmp read_config_x
read_config1:
	ldi r16,default_show_dly/8
	sts (w_time),r16
#ifdef default_random_mode
	sbi rnd_pics
#else
	cbi rnd_pics
#endif
	ldi ZL,LOW(cfile*2)
	ldi ZH,HIGH(cfile*2)
	cbi pname_in_ram
	call f_open
	tst r16
	brne read_config_x1		;use default settings if no config file found
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ldi r17,LOW(rd_buf_size)
	ldi r18,HIGH(rd_buf_size)
	call f_read_u
	tst r16
	breq read_config2
	cpi r16,FR_END_OF_FILE
	brne read_config_x1		;use default settings if something's wrong with the config file
read_config2:
	ldi ZL,LOW(timer_par*2)
	ldi ZH,HIGH(timer_par*2)
	rcall read_param
	breq read_config_x1		;use default settings in case of parameter error
	lsr r21
	ror r20
	lsr r21
	ror r20
	lsr r21
	ror r20
	brcc read_config4
read_config3:
	inc r20
read_config4:
	breq read_config3
	sts (w_time),r20
	ldi ZL,LOW(mode_par*2)
	ldi ZH,HIGH(mode_par*2)
	rcall read_param
	cbi rnd_pics
	breq read_config_x1
	sbi rnd_pics
read_config_x1:
	clr r16
read_config_x:
	ret

read_param:
	movw r15:r14,ZH:ZL
	ldi XL,LOW(sect_buf)
	ldi XH,HIGH(sect_buf)
	mov r24,r17
	mov r25,r18
	clr r20
	clr r21
read_param_lp:
	mov r16,r24
	or r16,r25
	breq read_param_x
	lpm r19,Z
	tst r19
	breq read_param_x1
	cpi r19,-1
	breq read_value
	ld r16,X
	cpi r16,'A'
	brcs read_param1
	cpi r16,'Z'+1
	brcc read_param1
	ori r16,$20
read_param1:
	cpi r19,' '
	brne read_param2
	cp r16,r19
	breq read_param10
	adiw Z,1
	rjmp read_param_lp
read_param2:
	cp r16,r19
	breq read_param9
	movw Z,r15:r14
	rjmp read_param10
read_param9:
	adiw Z,1
read_param10:
	adiw X,1
	sbiw r25:r24,1
	rjmp read_param_lp
read_param_x1:
	clz
read_param_x:
	ret

read_value:
	ld r16,X
	subi r16,'0'
	brcs read_value_x
	cpi r16,'9'+1
	brcc read_value_x
	ldi r19,10
	mul r19,r20
	ldi r19,0
	add r0,r16
	adc r1,r19
	movw r21:r20,r1:r0
	rjmp read_param10
read_value_x:
	adiw Z,1
	rjmp read_param_lp

timer_par: .db " timer = ",-1,0,-1
mode_par: .db " mode = random",0,-1

get_fname_seq:
	lds r17,(pict_file_ptr)
	lds r18,(pict_file_ptr+1)
	lds r19,(pict_file_ptr+2)
	lds r20,(pict_file_ptr+3)
	rcall f_lseek
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ldi r17,LOW(256)
	ldi r18,HIGH(256)
	rcall f_read
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ret

get_fname_rnd:
	ldi YL,LOW(fs)
	ldi YH,HIGH(fs)
	ldd r21,Y+fsize-fs
	ldd r22,Y+fsize-fs+1
	ldd r23,Y+fsize-fs+2
	ldd r24,Y+fsize-fs+3
	clr r4
	clr r5
	clr r6
	clr r7
get_fname_rnd_lp1:
	cp r4,r21
	cpc r5,r22
	cpc r6,r23
	cpc r7,r24
	brcc get_fname_rnd1
	adc r4,r4
	adc r5,r5
	adc r6,r6
	adc r7,r7
	brcc get_fname_rnd_lp1
get_fname_rnd1:
	ldi r25,5				;try 5 rounds max., otherwise use last result even if repeated
get_fname_rnd_rst:
	rcall rand_31
	ldi YL,LOW(random)
	ldi YH,HIGH(random)
	ld r17,Y
	ldd r18,Y+1
	ldd r19,Y+2
	ldd r20,Y+3
	and r17,r4
	and r18,r5
	and r19,r6
	and r20,r7
	cp r17,r21
	cpc r18,r22
	cpc r19,r23
	cpc r20,r24
	brcc get_fname_rnd_rst
	push r17
	push r18
	push r19
	push r20
	push r25
	rcall f_lseek
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ldi r17,LOW(512)
	ldi r18,HIGH(512)
	rcall f_read
	pop r25
	pop r24
	pop r23
	pop r22
	pop r21
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	mov r16,r21
	or r16,r22
	or r16,r23
	or r16,r24
	tst r16
	breq get_fname_rnd_chk
get_fname_rnd_lp2:
	mov r16,r17
	or r16,r18
	breq get_fname_rnd_x1
	ld r16,Z
	cpi r16,'/'
	breq get_fname_rnd2
	cpi r16,'\\'
	breq get_fname_rnd2
get_fname_rnd3:
	adiw Z,1
	subi r17,1
	sbci r18,0
	rjmp get_fname_rnd_lp2
get_fname_rnd2:
	ld r16,-Z
	adiw Z,1
	cpi ZH,HIGH(sect_buf)
	brcs get_fname_rnd_chk
	cpi r16,' '
	breq get_fname_rnd_chk
	cpi r16,10
	breq get_fname_rnd_chk
	cpi r16,13
	brne get_fname_rnd3
get_fname_rnd_chk:
	movw Y,Z
	subi YL,LOW(sect_buf)
	sbci YH,HIGH(sect_buf)
	clr r16
	add r21,YL
	adc r22,YH
	adc r23,r16
	adc r24,r16
	ldi r19,recent_entries
	ldi YL,LOW(recent_tab)
	ldi YH,HIGH(recent_tab)
get_fname_rnd_chk_lp:
	ld r16,Y
	cp r16,r21
	ldd r16,Y+1
	cpc r16,r22
	ldd r16,Y+2
	cpc r16,r23
	ldd r16,Y+3
	cpc r16,r24
	brne get_fname_rnd_chk1
	dec r25
get_fname_rnd_x1:
	breq get_fname_rnd_x		;exit after 5 times even if not found unrepeated file name
	ldi YL,LOW(fs)
	ldi YH,HIGH(fs)
	ldd r21,Y+fsize-fs
	ldd r22,Y+fsize-fs+1
	ldd r23,Y+fsize-fs+2
	ldd r24,Y+fsize-fs+3
	rjmp get_fname_rnd_rst
get_fname_rnd_chk1:
	adiw Y,4
	dec r19
	brne get_fname_rnd_chk_lp
	ldi YL,LOW(recent_tab)
	ldi YH,HIGH(recent_tab)
	ldi r19,(recent_entries-1)*4
get_fname_rnd_cp_lp:
	ldd r16,Y+4
	st Y+,r16
	dec r19
	brne get_fname_rnd_cp_lp
	st Y+,r21
	st Y+,r22
	st Y+,r23
	st Y,r24
get_fname_rnd_x:
	ret

check_optsensor:
	ldi r16,OPTSENSOR_MUX
	rcall adc_
	subi r16,LOW(OPTSENSOR_THR)
	sbci r17,HIGH(OPTSENSOR_THR)
	ret
adc_:
	push r16
	lds r16,(PRR0)
	andi r16,~(1<<PRADC)
	sts (PRR0),r16
	pop r16
	ori r16,1<<REFS0
	sts (ADMUX),r16
	ldi r16,1<<ADEN|1<<ADSC|1<<ADIF|1<<ADPS2|1<<ADPS1
	sts (ADCSRA),r16
adc_wt:
	lds r16,(ADCSRA)
	sbrs r16,ADIF
	rjmp adc_wt
	lds r16,(ADCL)
	lds r17,(ADCH)
	push r16
	clr r16				;disable ADC
	sts (ADCSRA),r16
	lds r16,(PRR0)
	ori r16,1<<PRADC
	sts (PRR0),r16
	pop r16
	ret

plane1:
	ldi r16,DATA_START_TRANSMISSION_1
	rcall ep_send_cmd
	ldi r24,0
	ldi r25,0
read_ram1_lp2:
	cbi SPI_SS
	ldi r16,3		;read command
	rcall rwspi1
	.if RAM_ADDR_BITS>16
	clr r16
	rcall rwspi1
	.endif
	mov r16,r25
	rcall rwspi1
	mov r16,r24
	rcall rwspi1
	ldi r17,200
	ldi ZL,LOW(buffer)
	ldi ZH,HIGH(buffer)
read_ram1_lp1:
	rcall rwspi1
	st Z+,r16
	dec r17
	brne read_ram1_lp1
	sbi SPI_SS
	ldi ZL,LOW(buffer)
	ldi ZH,HIGH(buffer)
	ldi r17,50
	sbrc r18,0
	rcall shuffle0
	sbrc r18,2
	rcall shuffle2
	subi r24,LOW(-200)
	sbci r25,HIGH(-200)
	ldi r16,HIGH(img_plane_size)
	cpi r24,LOW(img_plane_size)
	cpc r25,r16
	brne read_ram1_lp2
	add r18,r18
	ldi r16,DATA_START_TRANSMISSION_2
	rcall ep_send_cmd
	ldi r24,0
	ldi r25,0
read_ram2_lp2:
	cbi SPI_SS
	ldi r16,3		;read command
	rcall rwspi1
	.if RAM_ADDR_BITS>16
	clr r16
	rcall rwspi1
	.endif
	mov r16,r25
	rcall rwspi1
	mov r16,r24
	rcall rwspi1
	ldi r17,200
	ldi ZL,LOW(buffer)
	ldi ZH,HIGH(buffer)
read_ram2_lp1:
	rcall rwspi1
	st Z+,r16
	dec r17
	brne read_ram2_lp1
	sbi SPI_SS
	ldi ZL,LOW(buffer)
	ldi ZH,HIGH(buffer)
	ldi r17,50
	sbrc r18,1
	rcall shuffle1
	sbrc r18,3
	rcall shuffle3
	subi r24,LOW(-200)
	sbci r25,HIGH(-200)
	ldi r16,HIGH(img_plane_size)
	cpi r24,LOW(img_plane_size)
	cpc r25,r16
	brne read_ram2_lp2
	ret

shuffle0:
	ld r0,Z+
	bst r0,4
	bld r16,7
	bst r0,0
	bld r16,6
	ld r0,Z+
	bst r0,4
	bld r16,5
	bst r0,0
	bld r16,4
	ld r0,Z+
	bst r0,4
	bld r16,3
	bst r0,0
	bld r16,2
	ld r0,Z+
	bst r0,4
	bld r16,1
	bst r0,0
	bld r16,0
	rcall ep_send_data
	dec r17
	brne shuffle0
	ret

shuffle1:
	ld r0,Z+
	bst r0,5
	bld r16,7
	bst r0,1
	bld r16,6
	ld r0,Z+
	bst r0,5
	bld r16,5
	bst r0,1
	bld r16,4
	ld r0,Z+
	bst r0,5
	bld r16,3
	bst r0,1
	bld r16,2
	ld r0,Z+
	bst r0,5
	bld r16,1
	bst r0,1
	bld r16,0
	rcall ep_send_data
	dec r17
	brne shuffle1
	ret

shuffle2:
	ld r0,Z+
	bst r0,6
	bld r16,7
	bst r0,2
	bld r16,6
	ld r0,Z+
	bst r0,6
	bld r16,5
	bst r0,2
	bld r16,4
	ld r0,Z+
	bst r0,6
	bld r16,3
	bst r0,2
	bld r16,2
	ld r0,Z+
	bst r0,6
	bld r16,1
	bst r0,2
	bld r16,0
	rcall ep_send_data
	dec r17
	brne shuffle2
	ret

shuffle3:
	ld r0,Z+
	bst r0,7
	bld r16,7
	bst r0,3
	bld r16,6
	ld r0,Z+
	bst r0,7
	bld r16,5
	bst r0,3
	bld r16,4
	ld r0,Z+
	bst r0,7
	bld r16,3
	bst r0,3
	bld r16,2
	ld r0,Z+
	bst r0,7
	bld r16,1
	bst r0,3
	bld r16,0
	rcall ep_send_data
	dec r17
	brne shuffle3
	ret

ep_clr_ram:
	rcall ep_send_cmd
	ldi r24,LOW(15000)
	ldi r25,HIGH(15000)
ep_clr_ram_lp:
	ldi r16,-1
	rcall ep_send_data
	sbiw r25:r24,1
	brne ep_clr_ram_lp
	ret

ep_hndl_cmdstr:
ep_hndl_cmdstr_lp:
	lpm r17,Z+
	tst r17
	breq ep_hndl_cmdstr_x
	lpm r16,Z+
	cpi r16,WT_IDLE
	brne ep_hndl_cmdstr1
	rcall ep_wt_idle
	rjmp ep_hndl_cmdstr_lp
ep_hndl_cmdstr1:
	rcall ep_send_cmd
ep_hndl_data_lp:
	dec r17
	breq ep_hndl_cmdstr_lp
	lpm r16,Z+
	rcall ep_send_data
	rjmp ep_hndl_data_lp
ep_hndl_cmdstr_x:
	ret

ep_disp_frame:
	ldi r16,DISPLAY_REFRESH
	rcall ep_send_cmd
ep_wt_idle:
	cli
	sbic ep_bs
	rjmp ep_wt_idle_x
	cbi ep_rdy
	sei
ep_wt_idle1:
	sleep
	sbis ep_rdy
	rjmp ep_wt_idle1
ep_wt_idle_x:
	sei
	ret

ep_send_cmd:
	cbi ep_dc
	rjmp rwspi
ep_send_data:
	sbi ep_dc
rwspi:
	cbi ep_cs
	out SPDR,r16
rwspi_wt: in r16,SPSR
	sbrs r16,SPIF
	rjmp rwspi_wt
	in r16,SPDR
	sbi ep_cs
	ret

deselect:
	sbi CS
read_mmc_1st:
read_mmc:
	ldi r16,-1
send_mmc:
rwspi1:
	out SPDR,r16
rwspi1_wt: in r16,SPSR
	sbrs r16,SPIF
	rjmp rwspi1_wt
	in r16,SPDR
	ret

delay2ms:
	ldi r16,20
delay2ms_lp1:
	push r16
	clr r16
delay2ms_lp2:
	dec r16
	brne delay2ms_lp2
	pop r16
	dec r16
	brne delay2ms_lp1
	ret

; 31-bit random number 
; primitive polynomial order 31 
; (1)0100000 10101010 10010101 10110101 (20 AA 95 B5) 
; shift before XOR 
rand_31: 
	ldi YL,LOW(random)
	ldi YH,HIGH(random)
	ld r0,Y
	ldd r1,Y+1
	ldd r2,Y+2
	ldd r3,Y+3
	lsl r0			; shift first 
	rol r1 
	rol r2 
	rol r3 
	sbrs r3,7		; test MSB 
	rjmp rand_store	; clear, no XOR 
	ldi r16,0xB5	; first poly byte 
	eor r0,r16		; XOR it in 
	ldi r16,0x95	; second byte of poly 
	eor r1,r16
	ldi r16,0xAA	; same routine 
	eor r2,r16
	ldi r16,0x20
	eor r3,r16
rand_store: 
	st Y,r0
	std Y+1,r1
	std Y+2,r2
	std Y+3,r3
	ret

;decrement bit counter and feed next byte of huffman coded image data into the JPEG decoder
;capable to skip restart markers, but not other marker segments
;input:	r16			bit stream buffer
;		r17			count of remaining bits
;output: r16		adapted bit stream buffer
;		r17			adapted count of remaining bits
	.macro get_nxt_byte
	dec r17										;decrement bit stream counter
	brne get_nxt_bytex							;skip if still bits left
	ldi r17,8
get_nxt_byte_rld: rd_byte						;get next data byte
	cpi r16,$ff									;possible start of marker?
	brne get_nxt_bytex							;no; deliver data byte
	rd_byte										;get next data byte
	tst r16										;2nd byte==0 for byte stuffing?
	brne get_nxt_byte_rld						;no, assume restart marker and get next byte 
	ldi r16,$ff									;deliver original value
get_nxt_bytex:
	.endmacro

;general read macro, delivers single byte input to JPEG decoder
;output: r16		byte read from source
	.macro rd_byte
	rcall rd_byte1
	.endmacro

rd_byte1:
	sbis rd_src
	rjmp rd_byte_flash
rd_byte_sect:
	push ZL
	push ZH
	lds ZL,(rd_ptr)
	lds ZH,(rd_ptr+1)
	cpi ZH,HIGH(sect_buf+rd_buf_size)
	brne rd_byte_sect1
	rcall rd_byte_rd_sect
rd_byte_sect1:
	ld r16,Z+
	sts (rd_ptr),ZL
	sts (rd_ptr+1),ZH
	pop ZH
	pop ZL
	ret

rd_byte_rd_sect:
	push r17
	push r18
	push r19
	push r20
	push r21
	push XL
	push XH
	push YL
	push YH
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ldi r17,LOW(rd_buf_size)
	ldi r18,HIGH(rd_buf_size)
	cbi CS
	rcall f_read_u
	sbi CS
	pop YH
	pop YL
	pop XH
	pop XL
	pop r21
	pop r20
	pop r19
	pop r18
	pop r17
	ldi ZL,LOW(sect_buf)
	ldi ZH,HIGH(sect_buf)
	ret

;get single byte from controller's flash memory
;capable to handle memory addresses >65535
;output: r16		byte read from flash
rd_byte_flash:
	push ZL
	push ZH
	lds ZL,(rd_ptr)
	lds ZH,(rd_ptr+1)
	lpm r16,Z+
	sts (rd_ptr),ZL
	sts (rd_ptr+1),ZH
	pop ZH
	pop ZL
	ret

;JPEG decoder, decodes image and shows it on display
;first part, preparation of the tables and basic values
dec_jpg: ldi r16,LOW(ht)
	ldi r17,HIGH(ht)
	movw r15:r14,r17:r16						;r15:r14 hold start address of Huffman tree to be built
file_scan_lp: rd_byte
	cpi r16,$ff									;check if 1st byte of marker
	breq file_scan
	ldi r16,0
	ret
file_error: 
	ldi r16,-1
	ret

file_scan: rd_byte
	cpi r16,$d8									;SOI
	breq file_scan_lp
	mov r17,r16
	rd_byte										;read segment length parameter
	mov r25,r16
	rd_byte
	mov r24,r16
	sbiw r25:r24,2
	cpi r17,$db									;DQT
	brne file_scan1
dqt_lp: rd_byte
	tst r16										;only 8 bit precision supported and table index 0 (Luminance) supported
	brne file_error
	ldi YL,LOW(q_tbl)
	ldi YH,HIGH(q_tbl)
	ldi r18,64									;read 64 quantization values
dqt_cp: rd_byte
	st Y+,r16
	dec r18
	brne dqt_cp
	subi r24,65
	brne file_error
	rjmp file_scan_lp

file_scan1: cpi r17,$c4							;DHT
	brne file_scan2
	movw YH:YL,r15:r14
	clr r6
dht_scan: rcall build_ht						;build one huffman deconding tree
	tst r16
	brne file_error
	mov r16,r24
	or r16,r25
	brne dht_scan								;continue if more data left
	movw r15:r14,YH:YL
	rjmp file_scan_lp
file_error_jp: rjmp file_error

file_scan3: cpi r17,$da							;SOS
	brne file_skip
	rjmp decode
file_skip: rd_byte								;skip any other segment
	sbiw r25:r24,1
	brne file_skip
	rjmp file_scan_lp

file_scan2: cpi r17,$c0							;SOF0
	brne file_scan3
	rd_byte
	cpi r16,8									;only 8 bit data precision supported
file_error2: brne file_error_jp
	rd_byte
	mov r17,r16
	rd_byte
	subi r16,LOW(y_)
	sbci r17,HIGH(y_)
	brne file_error_jp
	rd_byte
	mov r17,r16
	rd_byte
	subi r16,LOW(x_)
	sbci r17,HIGH(x_)
	brne file_error_jp
	rd_byte
	cpi r16,1									;only single component (Y) supported
	brne file_error_jp
	subi r24,6
	rd_byte
	mov r17,r16									;component ID
	rd_byte
	cpi r16,$11									;only grayscale with sampling factors 1x1 supported
	brne file_error_jp
	rd_byte
	rjmp file_scan_lp

;decodes huffman coded image data and shows it on display
decode:

;skip component information preceding image data assuming components are assigned the usual way
skip_sos_hd: rd_byte
	sbiw r25:r24,1
	brne skip_sos_hd
;DC value per component is coded as difference to previous block
;initialize with zero for Y
	sts (prv_dc),r2
	sts (prv_dc+1),r2
	ldi r17,1

	clr r25										;line block counter, one block = 8 pixel lines
scan_loop: 
	clr r24										;column block counter, one block = 8 pixel columns
scan_loop1:
	sts (x_ptr),r24
	sts (y_ptr),r25
;read data for one MCU
	clr r2
	clr r19
	ldi r22,LOW(prv_dc)
	ldi r23,HIGH(prv_dc)
	ldi YL,LOW(blk_y)
	ldi YH,HIGH(blk_y)
	rcall get_blk								;read Y block

	push r16
	push r17
	push ZL
	push ZH

	ldi YL,LOW(blk_y)
	ldi YH,HIGH(blk_y)
	rcall idct									;transform Y block

	lds r24,(x_ptr)
	lds r25,(y_ptr)

	ldi YL,LOW(blk_y)
	ldi YH,HIGH(blk_y)
	
	ldi r20,x_/2
	mul r20,r25
	add r0,r0
	adc r1,r1
	add r0,r0
	adc r1,r1
	add r0,r0
	adc r1,r1
	movw r19:r18,r1:r0
	ldi r20,4
	mul r20,r24	
	add r18,r0
	adc r19,r1

	ldi r20,8
	mul r20,r25
	ldi r16,LOW(y_-8)
	ldi r17,HIGH(y_-8)
	sub r16,r0
	sbc r17,r1
	brpl store_line
	add r20,r16
store_line:
	cbi SPI_SS
	ldi r16,2		;SPI RAM write command
	rcall rwspi1
	.if RAM_ADDR_BITS>16
	clr r16
	rcall rwspi1
	.endif
	mov r16,r19
	rcall rwspi1
	mov r16,r18
	rcall rwspi1
	ld r17,Y
	andi r17,$f0
	ldd r16,Y+2
	andi r16,$f0
	swap r16
	or r16,r17
	rcall rwspi1 
	ldd r17,Y+4
	andi r17,$f0
	ldd r16,Y+6
	andi r16,$f0
	swap r16
	or r16,r17
	rcall rwspi1 
	ldd r17,Y+8
	andi r17,$f0
	ldd r16,Y+10
	andi r16,$f0
	swap r16
	or r16,r17
	rcall rwspi1 
	ldd r17,Y+12
	andi r17,$f0
	ldd r16,Y+14
	andi r16,$f0
	swap r16
	or r16,r17
	rcall rwspi1 
	sbi SPI_SS
	adiw Y,16
	subi r18,LOW(-x_/2)
	sbci r19,HIGH(-x_/2)
	dec r20
	brne store_line
	pop ZH
	pop ZL
	pop r17
	pop r16

	inc r24
	cpi r24,x_/8
	breq scan1
	rjmp scan_loop1
scan1:
	inc r25
	cpi r25,(y_+7)/8
	breq scan_x
	rjmp scan_loop
scan_x: 
	ldi r16,0
	ret


;prepare one block (8 x 8 values) to be fed into IDCT calculation
;including decompression, dequantization, scaling and re-ordering
;
;input:	r19			pointer offset to corresponding Huffman tree pair
;		r23:r22		pointer to stored DC value of previous block
;		YH:YL		pointer to component block storage
get_blk:
	ldi r20,LOW(q_tbl)
	ldi r21,HIGH(q_tbl)
	movw r5:r4,r21:r20							;r5:r4 holds address of quantization table
	push YL
	push YH
	ldi r24,LOW(pow2tab*2-2)
	ldi r25,HIGH(pow2tab*2-2)
	movw r11:r10,r25:r24						;r11:r10 points to a table holding power of 2 values for decoding AC/DC values

	ldi YL,LOW(ht_ptr)
	ldi YH,HIGH(ht_ptr)
	add YL,r19
	adc YH,r2
	ld r24,Y+
	ld r25,Y+									;r25:r24 holds the address of the appropriate Huffman decoding tree for DC
	push YL
	push YH

;clear temporary output buffer at once so that empty ranges can be skipped during decompression instead of clearing them individually
	ldi YL,LOW(tmp_buf)
	ldi YH,HIGH(tmp_buf)
	movw XH:XL,YH:YL							;XH:XL holds pointer to temporary output buffer
	ldi r20,128/32
clr_blk_lp: 
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	st Y+,r2
	dec r20
	brne clr_blk_lp

	rcall get_node								;find corresponding node entry for DC value
	rcall get_value								;get remaining bits and convert to signed value
	movw YH:YL,r23:r22
	ld r18,Y									;get previous DC value
	ldd r19,Y+1
	add r20,r18									;actual DC value is difference between previous value and the one just read
	adc r21,r19
	st Y,r20									;remember for next time
	std Y+1,r21
	movw YH:YL,r5:r4
	ld r19,Y+									;get quantization value from table
	mulsu r21,r19								;apply dequantization
	mov r3,r0
	mul r20,r19
	add r1,r3
	movw r5:r4,YH:YL
	st X+,r0
	st X+,r1

	ldi r18,63
	mov r12,r18
	ldi r24,LOW(arai_scale_zz*2)
	ldi r25,HIGH(arai_scale_zz*2)
	movw r7:r6,r25:r24							;r7:r6 points to current entry in table of Arai scaling factors
	pop YH
	pop YL
	ld r24,Y+
	ld r25,Y									;r25:r24 now holds the address of the appropriate Huffman decoding tree for AC
get_ac_lp:
	rcall get_node								;find corresponding node entry for AC value
	tst r19										;EOB node found?
	breq get_acx								;yes, decompression for this block is finished
	rcall get_value								;get remaining bits and convert to signed value, skipping range of zero coded values
	movw YH:YL,r5:r4
	ld r19,Y+									;get quantization value from table
	mulsu r21,r19								;apply dequantization
	mov r3,r0
	mul r20,r19
	add r1,r3
	movw r21:r20,r1:r0
	movw r5:r4,YH:YL

	movw ZH:ZL,r7:r6
	lpm r18,Z+									;get Arai scaling factor
	lpm r19,Z+
	movw r7:r6,ZH:ZL

;apply prescaling according to Arai's IDCT algorithm
;	alternatively the quantization tables could be prepared accordingly once before,
;	saving the need of an extra multiplication here; it might slightly increase decoding speed
;
;	but to retain accuracy with regards to memory (RAM) consumption it's done this way
	fmulsu r21,r19
	movw r9:r8,r1:r0
	fmul r20,r18
	adc r8,r2
	mov r3,r1
	fmulsu r21,r18
	sbc r9,r2
	add r3,r0
	adc r8,r1
	adc r9,r2
	fmul r20,r19
	adc r9,r2
	add r3,r0
	adc r8,r1
	adc r9,r2
	movw r21:r20,r9:r8

	st X+,r20
	st X+,r21
	dec r12
	brne get_ac_lp
get_acx:
	pop XH
	pop XL										;XH:XL points to component block storage

;copy values from temporary buffer to component block, reverting the zigzag order
	ldi YL,LOW(tmp_buf)
	ldi YH,HIGH(tmp_buf)
	ldi ZL,LOW(tmp_buf+64)
	ldi ZH,HIGH(tmp_buf+64)
;0
	ld r18,Y
	ldd r19,Y+1
	st X+,r18
	st X+,r19
;1
	ldd r18,Y+2
	ldd r19,Y+3
	st X+,r18
	st X+,r19
;5
	ldd r18,Y+10
	ldd r19,Y+11
	st X+,r18
	st X+,r19
;6
	ldd r18,Y+12
	ldd r19,Y+13
	st X+,r18
	st X+,r19
;14
	ldd r18,Y+28
	ldd r19,Y+29
	st X+,r18
	st X+,r19
;15
	ldd r18,Y+30
	ldd r19,Y+31
	st X+,r18
	st X+,r19
;27
	ldd r18,Y+54
	ldd r19,Y+55
	st X+,r18
	st X+,r19
;28
	ldd r18,Y+56
	ldd r19,Y+57
	st X+,r18
	st X+,r19
;2
	ldd r18,Y+4
	ldd r19,Y+5
	st X+,r18
	st X+,r19
;4
	ldd r18,Y+8
	ldd r19,Y+9
	st X+,r18
	st X+,r19
;7
	ldd r18,Y+14
	ldd r19,Y+15
	st X+,r18
	st X+,r19
;13
	ldd r18,Y+26
	ldd r19,Y+27
	st X+,r18
	st X+,r19
;16
	ldd r18,Y+32
	ldd r19,Y+33
	st X+,r18
	st X+,r19
;26
	ldd r18,Y+52
	ldd r19,Y+53
	st X+,r18
	st X+,r19
;29
	ldd r18,Y+58
	ldd r19,Y+59
	st X+,r18
	st X+,r19
;42
	ldd r18,Z+20
	ldd r19,Z+21
	st X+,r18
	st X+,r19
;3
	ldd r18,Y+6
	ldd r19,Y+7
	st X+,r18
	st X+,r19
;8
	ldd r18,Y+16
	ldd r19,Y+17
	st X+,r18
	st X+,r19
;12
	ldd r18,Y+24
	ldd r19,Y+25
	st X+,r18
	st X+,r19
;17
	ldd r18,Y+34
	ldd r19,Y+35
	st X+,r18
	st X+,r19
;25
	ldd r18,Y+50
	ldd r19,Y+51
	st X+,r18
	st X+,r19
;30
	ldd r18,Y+60
	ldd r19,Y+61
	st X+,r18
	st X+,r19
;41
	ldd r18,Z+18
	ldd r19,Z+19
	st X+,r18
	st X+,r19
;43
	ldd r18,Z+22
	ldd r19,Z+23
	st X+,r18
	st X+,r19
;9
	ldd r18,Y+18
	ldd r19,Y+19
	st X+,r18
	st X+,r19
;11
	ldd r18,Y+22
	ldd r19,Y+23
	st X+,r18
	st X+,r19
;18
	ldd r18,Y+36
	ldd r19,Y+37
	st X+,r18
	st X+,r19
;24
	ldd r18,Y+48
	ldd r19,Y+49
	st X+,r18
	st X+,r19
;31
	ldd r18,Y+62
	ldd r19,Y+63
	st X+,r18
	st X+,r19
;40
	ldd r18,Z+16
	ldd r19,Z+17
	st X+,r18
	st X+,r19
;44
	ldd r18,Z+24
	ldd r19,Z+25
	st X+,r18
	st X+,r19
;53
	ldd r18,Z+42
	ldd r19,Z+43
	st X+,r18
	st X+,r19
;10
	ldd r18,Y+20
	ldd r19,Y+21
	st X+,r18
	st X+,r19
;19
	ldd r18,Y+38
	ldd r19,Y+39
	st X+,r18
	st X+,r19
;23
	ldd r18,Y+46
	ldd r19,Y+47
	st X+,r18
	st X+,r19
;32
	ld r18,Z
	ldd r19,Z+1
	st X+,r18
	st X+,r19
;39
	ldd r18,Z+14
	ldd r19,Z+15
	st X+,r18
	st X+,r19
;45
	ldd r18,Z+26
	ldd r19,Z+27
	st X+,r18
	st X+,r19
;52
	ldd r18,Z+40
	ldd r19,Z+41
	st X+,r18
	st X+,r19
;54
	ldd r18,Z+44
	ldd r19,Z+45
	st X+,r18
	st X+,r19
;20
	ldd r18,Y+40
	ldd r19,Y+41
	st X+,r18
	st X+,r19
;22
	ldd r18,Y+44
	ldd r19,Y+45
	st X+,r18
	st X+,r19
;33
	ldd r18,Z+2
	ldd r19,Z+3
	st X+,r18
	st X+,r19
;38
	ldd r18,Z+12
	ldd r19,Z+13
	st X+,r18
	st X+,r19
;46
	ldd r18,Z+28
	ldd r19,Z+29
	st X+,r18
	st X+,r19
;51
	ldd r18,Z+38
	ldd r19,Z+39
	st X+,r18
	st X+,r19
;55
	ldd r18,Z+46
	ldd r19,Z+47
	st X+,r18
	st X+,r19
;60
	ldd r18,Z+56
	ldd r19,Z+57
	st X+,r18
	st X+,r19
;21
	ldd r18,Y+42
	ldd r19,Y+43
	st X+,r18
	st X+,r19
;34
	ldd r18,Z+4
	ldd r19,Z+5
	st X+,r18
	st X+,r19
;37
	ldd r18,Z+10
	ldd r19,Z+11
	st X+,r18
	st X+,r19
;47
	ldd r18,Z+30
	ldd r19,Z+31
	st X+,r18
	st X+,r19
;50
	ldd r18,Z+36
	ldd r19,Z+37
	st X+,r18
	st X+,r19
;56
	ldd r18,Z+48
	ldd r19,Z+49
	st X+,r18
	st X+,r19
;59
	ldd r18,Z+54
	ldd r19,Z+55
	st X+,r18
	st X+,r19
;61
	ldd r18,Z+58
	ldd r19,Z+59
	st X+,r18
	st X+,r19
;35
	ldd r18,Z+6
	ldd r19,Z+7
	st X+,r18
	st X+,r19
;36
	ldd r18,Z+8
	ldd r19,Z+9
	st X+,r18
	st X+,r19
;48
	ldd r18,Z+32
	ldd r19,Z+33
	st X+,r18
	st X+,r19
;49
	ldd r18,Z+34
	ldd r19,Z+35
	st X+,r18
	st X+,r19	
;57
	ldd r18,Z+50
	ldd r19,Z+51
	st X+,r18
	st X+,r19
;58
	ldd r18,Z+52
	ldd r19,Z+53
	st X+,r18
	st X+,r19
;62
	ldd r18,Z+60
	ldd r19,Z+61
	st X+,r18
	st X+,r19
;63
	ldd r18,Z+62
	ldd r19,Z+63
	st X+,r18
	st X+,r19
	ret

;deliver actual value based on previously determined code entry from Huffman decoding tree
;skip range of zero coded values
;input:	r5:r4		pointer to current entry in quantization table
;		r7:r6		pointer to current entry in Arai scaling factor table
;		r11:r10		pointer to power of 2 table
;		r12			number of remaining values in current block
;		r16			bit stream buffer
;		r17			number of valid bits in bit stream buffer
;		r19			code entry from Huffman decoding tree
;		XH:XL		pointer to current position in temporary output buffer
;output: r5:r4		adapted pointer to current entry in quantization table
;		r7:r6		adapted pointer to current entry in Arai scaling factor table
;		r12			adapted number of remaining values in current block
;		r16			adapted bit stream buffer
;		r17			adapted number of valid bits in bit stream buffer
;		r21:r20		signed AC/DC value
;		XH:XL		adapted pointer to current position in temporary output buffer
get_value: clr r20
	clr r21
	ldi r18,$f0
	and r18,r19
	breq get_value1								;no range of zero values to skip
	swap r18									;r18 holds number of zero values to skip
	sub r12,r18
	add r4,r18
	adc r5,r2
	add r18,r18
	add XL,r18
	adc XH,r2
	add r6,r18
	adc r7,r2
get_value1: andi r19,15
	breq get_valuex								;skip if bit count of output value is zero
	mov r3,r19
get_val_lp: get_nxt_byte
	lsl r16
	rol r20
	rol r21
	dec r19
	brne get_val_lp
	movw ZH:ZL,r11:r10
	add r3,r3
	add ZL,r3
	adc ZH,r2
	lpm r18,Z+									;get power of 2 value for bit count
	lpm r19,Z
	movw r1:r0,r19:r18
	and r0,r20
	and r1,r21
	or r0,r1
	brne get_valuex								;skip if positive value
	add r18,r18									;invert for negative value
	adc r19,r19
	subi r18,1
	sbci r19,0
	com r19
	neg r18
	sbci r19,-1
	add r20,r18
	adc r21,r19
get_valuex: ret

;read bits from input stream, find corresponding node in Huffman decoding tree
;input:	r16			bit stream buffer
;		r17			number of remaining bits in bit stream buffer
;		r25:r24		address of Huffman decoding tree
;output: r16		adapted bit stream buffer
;		r17			adapted number of remaining bits in bit stream buffer
;		r19			code entry from Huffman decoding tree
get_node: movw YH:YL,r25:r24
	ld r18,Y+									;get distance to next tree level
get_node_lp: get_nxt_byte
	lsl r16
	brcc get_node2								;skip if current bit was 0
	subi r18,-2									;add two if current bit was set
get_node2: add YL,r18							;add distance to pointer
	adc YH,r2
	ld r18,Y+									;get current distance
	tst r18
	brne get_node_lp							;repeat if distance >0
	ld r19,Y									;get code value
	ret

;construct Huffman decoding tree
;tree is built of 2-byte nodes
;1st byte !=0: branch node, forward distance -1 to next level if current bit is 0, add two if current bit is 1
;         ==0: leaf node
;2nd byte    : leaf node value/code entry
;input:	YH:YL		start address of Huffman decoding tree to be built
;output: YH:YL		next available address for Huffman decoding tree
build_ht: rd_byte								;get Huffman decoding tree information
	sbiw r25:r24,1								;segment length count -1
	ldi r17,$10
	and r17,r16									;bit 4 carries type of Huffman decoding tree, 0: DC table, 1: AC table
	swap r17
	andi r16,15									;bit 0...3 carry number of Huffman decoding tree, only 0 or 1 supported
	add r16,r16									;construct pointer table index
	add r16,r17
	add r16,r16
	ldi XL,LOW(ht_ptr)
	ldi XH,HIGH(ht_ptr)
	add XL,r16
	adc XH,r6
	st X+,YL									;store start address of current Huffman decoding tree in pointer table
	st X,YH
	movw r5:r4,YH:YL
	ldi XL,LOW(tmp_buf)
	ldi XH,HIGH(tmp_buf)
	ldi r17,16
get_symcnt: rd_byte								;get symbol count for corresponding code length
	st X+,r16
	dec r17
	brne get_symcnt
	ldi r17,16
push_symcnt: ld r16,-X
	push r16									;push symbol count onto stack
	dec r17
	brne push_symcnt
	sbiw r25:r24,16								;subtract 16 from segment length count
	ldi r19,16									;code word level count
	st Y,r6										;create root node
	adiw Y,2									;new end of Huffman decoding tree
	clr r22										;r23:r22 current code word
	clr r23
	ldi r20,1									;r21:r20 current code bit mask
	clr r21
bld_htlp: pop r17								;get symbol count for current code word level
	tst r17
	breq bld_htn								;skip if no symbols at this level
bld_htlp1: rd_byte								;get code value
	sbiw r25:r24,1								;segment length count -1
	rcall add_node								;add node(s) and store code value into leaf node
	cpi YL,LOW(ht+ht_size)
	ldi r16,HIGH(ht+ht_size)
	cpc YH,r16
	ldi r16,-1
	brcc build_htx
	subi r22,-1									;increment code word
	sbci r23,-1
	dec r17										;decrement symbol count
	brne bld_htlp1
bld_htn:
	add r22,r22									;advance to next code word level (up to 16 bit)
	adc r23,r23
	add r20,r20
	adc r21,r21
	dec r19										;decrement level count
	brne bld_htlp
	ldi r16,0
build_htx:
	ret

;add new node(s) to Huffman decoding tree if not already existing and store code value into leaf node
;input:	r5:r4		start address of current Huffman decoding tree
;		r16			code value
;		r21:r20		code bit mask
;		r23:r22		current code word
;		YH:YL		next available address after current Huffman decoding tree
add_node: 
	movw r1:r0,r21:r20							;r1:r0 holds current code mask upper bit
	movw XH:XL,r5:r4							;XH:XL points to start address of current Huffman decoding tree
add_node_lp: ld r2,X							;get distance to next level
	tst r2
	brne add_node1								;skip if distance already set
	mov r2,YL
	sub r2,XL									;calculate distance to current end of Huffman decoding tree
	dec r2
	st X,r2										;store distance on current node
	st Y,r6										;create new node for code bit 0
	std Y+2,r6									;create new node for code bit 1
	adiw Y,4									;adapt pointer to end of current Huffman decoding tree
add_node1: movw r3:r2,r1:r0						;get current code bit mask
	and r2,r22									;mask out other bits of current code word
	and r3,r23
	or r2,r3									;check if current bit in code word is set
	ld r18,X+									;get distance to next level
	breq add_node2
	subi r18,-2									;add two if bit in code word was set
add_node2: add XL,r18							;add distance to pointer
	adc XH,r6
	lsr r1										;shift code bit mask 
	ror r0
	brcc add_node_lp							;repeat until all bits were checked
	adiw X,1
	st X,r16									;store code value
	ret

;Inverse Discrete Cosine Transform
;implementation based on Tiny JPEG Decompressor
;input:	YH:YL		pointer to component block storage

.equ m13=181	;32768*cos(4*pi/16),					23170>>7=181
.equ m2=139		;32768*(cos(2*pi/16)-cos(6*pi/16)),		17733>>7=139
.equ m4=167		;32768*(cos(2*pi/16)+cos(6*pi/16)),		42813>>8=167
.equ m5=237		;32768*cos(2*pi/16),					30274>>7=237
idct: movw ZH:ZL,YH:YL
	subi ZL,LOW(-64)
	sbci ZH,HIGH(-64)
	push YL
	push YH

	ldi r19,8

idct_col_lp:
	ld r8,Y					;v0
	ldd r9,Y+1
	ldd r10,Y+32			;v1
	ldd r11,Y+33
	ld r12,Z				;v2
	ldd r13,Z+1
	ldd r14,Z+32			;v3
	ldd r15,Z+33

	movw r17:r16,r11:r10	;t11=(v1-v3)*m13
	sub r16,r14
	sbc r17,r15
 	ldi r18,m13
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2

	movw r27:r26,r9:r8		;t10=v0+v2
	add r26,r12
	adc r27,r13

	movw r25:r24,r9:r8		;t12=v0-v2
	sub r24,r12
	sbc r25,r13

	add r14,r10				;v3+=v1
	adc r15,r11

	sub r0,r14				;t11-=v3
	sbc r1,r15

	movw r9:r8,r27:r26		;v0=t10+v3
	add r8,r14
	adc r9,r15
	
	sub r26,r14				;(v3)=t10-v3
	sbc r27,r15

	movw r23:r22,r1:r0		;v1=t11+t12
	add r22,r24
	adc r23,r25

	sub r24,r0				;(v2)=t12-t11
	sbc r25,r1

	ldd r12,Z+48			;v4
	ldd r13,Z+49
	ldd r14,Y+16			;v5
	ldd r15,Y+17

	movw r7:r6,r15:r14		;t10=v5-v4
	sub r6,r12
	sbc r7,r13

	movw r5:r4,r15:r14		;t11=v5+v4
	add r4,r12
	adc r5,r13

	ldd r12,Z+16			;v6
	ldd r13,Z+17
	ldd r14,Y+48			;v7
	ldd r15,Y+49

	movw r21:r20,r13:r12	;t12=v6-v7
	sub r20,r14
	sbc r21,r15

	add r14,r12				;v7+=v6
	adc r15,r13

	movw r17:r16,r5:r4		;v5=(t11-v7)*m13
	sub r16,r14
	sbc r17,r15
	ldi r18,m13
	clr r11
	fmul r16,r18
	adc r11,r11
	mov r10,r1
	fmulsu r17,r18
	add r10,r0
	adc r11,r1

	add r14,r4				;v7+=t11
	adc r15,r5

	movw r17:r16,r7:r6		;t13=(t10+t12)*m5
	add r16,r20
	adc r17,r21
	ldi r18,m5
	clr r5
	fmul r16,r18
	adc r5,r5
	mov r4,r1
	fmulsu r17,r18
	add r4,r0
	adc r5,r1

	movw r17:r16,r7:r6		;v4=t13-t10*m2
	ldi r18,m2
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2
	movw r13:r12,r5:r4
	sub r12,r0
	sbc r13,r1

	ldi r18,m4				;(v6)=t13-t12*m4-v7
	clr r2
	fmul r20,r18
	adc r2,r2
	mov r20,r0
	mov r3,r1
	fmulsu r21,r18
	add r0,r3
	adc r1,r2
	add r20,r20
	adc r0,r0
	adc r1,r1
	sub r4,r0
	sbc r5,r1
	sub r4,r14
	sbc r5,r15

	sub r10,r4				;v5-=v6
	sbc r11,r5

	sub r12,r10				;v4-=v5
	sbc r13,r11

	movw r1:r0,r9:r8		;v0+v7
	add r0,r14
	adc r1,r15
	st Y+,r0
	st Y+,r1

	sub r8,r14				;v0-v7
	sbc r9,r15
	std Z+48,r8
	std Z+49,r9

	movw r1:r0,r23:r22		;v1+v6
	add r0,r4
	adc r1,r5
	std Y+14,r0
	std Y+15,r1

	sub r22,r4				;v1-v6
	sbc r23,r5
	std Z+32,r22
	std Z+33,r23

	movw r1:r0,r25:r24		;v2+v5
	add r0,r10
	adc r1,r11
	std Y+30,r0
	std Y+31,r1

	sub r24,r10				;v2+v5
	sbc r25,r11
	std Z+16,r24
	std Z+17,r25

	movw r1:r0,r27:r26		;v3+v4
	add r0,r12
	adc r1,r13
	std Y+46,r0
	std Y+47,r1

	sub r26,r12				;v3-v4
	sbc r27,r13
	st Z+,r26
	st Z+,r27

	dec r19
	breq idct1
	rjmp idct_col_lp
idct1:

	pop YH
	pop YL
	ldi r19,8

idct_row_lp:
	ld r20,Y				;v0
	ldd r21,Y+1
	ldd r10,Y+4				;v1
	ldd r11,Y+5
	ldd r12,Y+8				;v2
	ldd r13,Y+9
	ldd r14,Y+12			;v3
	ldd r15,Y+13

	subi r20,LOW(-$400)		;v0+=-$400
	sbci r21,HIGH(-$400)

	movw r17:r16,r11:r10	;t11=(v1-v3)*m13
	sub r16,r14
	sbc r17,r15
 	ldi r18,m13
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2

	movw r27:r26,r21:r20	;t10=v0+v2
	add r26,r12
	adc r27,r13

	movw r25:r24,r21:r20	;t12=v0-v2
	sub r24,r12
	sbc r25,r13

	add r14,r10				;v3+=v1
	adc r15,r11

	sub r0,r14				;t11-=v3
	sbc r1,r15

	movw r31:r30,r27:r26	;v0=t10+v3
	add r30,r14
	adc r31,r15
	
	sub r26,r14				;(v3)=t10-v3
	sbc r27,r15

	movw r23:r22,r1:r0		;v1=t11+t12
	add r22,r24
	adc r23,r25

	sub r24,r0				;(v2)=t12-t11
	sbc r25,r1

	ldd r8,Y+14				;v4
	ldd r9,Y+15
	ldd r10,Y+2				;v5
	ldd r11,Y+3
	ldd r12,Y+10			;v6
	ldd r13,Y+11
	ldd r14,Y+6				;v7
	ldd r15,Y+7

	movw r7:r6,r11:r10		;t10=v5-v4
	sub r6,r8
	sbc r7,r9

	movw r5:r4,r11:r10		;t11=v5+v4
	add r4,r8
	adc r5,r9

	movw r21:r20,r13:r12	;t12=v6-v7
	sub r20,r14
	sbc r21,r15

	add r14,r12				;v7+=v6
	adc r15,r13

	movw r17:r16,r5:r4		;v5=(t11-v7)*m13
	sub r16,r14
	sbc r17,r15
	ldi r18,m13
	clr r11
	fmul r16,r18
	adc r11,r11
	mov r10,r1
	fmulsu r17,r18
	add r10,r0
	adc r11,r1

	add r14,r4				;v7+=t11
	adc r15,r5

	movw r17:r16,r7:r6		;t13=(t10+t12)*m5
	add r16,r20
	adc r17,r21
	ldi r18,m5
	clr r5
	fmul r16,r18
	adc r5,r5
	mov r4,r1
	fmulsu r17,r18
	add r4,r0
	adc r5,r1

	movw r17:r16,r7:r6		;v4=t13-t10*m2
	ldi r18,m2
	clr r2
	fmul r16,r18
	adc r2,r2
	mov r3,r1
	fmulsu r17,r18
	add r0,r3
	adc r1,r2
	movw r9:r8,r5:r4
	sub r8,r0
	sbc r9,r1

	ldi r18,m4				;(v6)=t13-t12*m4-v7
	clr r2
	fmul r20,r18
	adc r2,r2
	mov r20,r0
	mov r3,r1
	fmulsu r21,r18
	add r0,r3
	adc r1,r2
	add r20,r20
	adc r0,r0
	adc r1,r1
	sub r4,r0
	sbc r5,r1
	sub r4,r14
	sbc r5,r15

	sub r10,r4				;v5-=v6
	sbc r11,r5

	sub r8,r10				;v4-=v5
	sbc r9,r11

	.macro byteclip 
	brpl bc1
	clr @0
	rjmp bc3
bc1: lsr @1
	ror @0
	lsr @1
	ror @0
	lsr @1
	breq bc2
	ldi @0,-1
	rjmp bc3
bc2: ror @0
bc3:
	.endmacro

	movw r21:r20,r31:r30	;v0+v7
	add r20,r14
	adc r21,r15
	byteclip r20,r21
	st Y,r20

	sub r30,r14				;v0-v7
	sbc r31,r15
	byteclip r30,r31
	std Y+14,r30

	movw r21:r20,r23:r22	;v1+v6
	add r20,r4
	adc r21,r5
	byteclip r20,r21
	std Y+2,r20

	sub r22,r4				;v1-v6
	sbc r23,r5
	byteclip r22,r23
	std Y+12,r22

	movw r21:r20,r25:r24	;v2+v5
	add r20,r10
	adc r21,r11
	byteclip r20,r21
	std Y+4,r20

	sub r24,r10				;v2-v5
	sbc r25,r11
	byteclip r24,r25
	std Y+10,r24

	movw r21:r20,r27:r26	;v3+v4
	add r20,r8
	adc r21,r9
	byteclip r20,r21
	std Y+6,r20

	sub r26,r8				;v3-v4
	sbc r27,r9
	byteclip r26,r27
	std Y+8,r26

	adiw Y,16
	dec r19
	breq idct_x
	rjmp idct_row_lp
idct_x: ret


f_lseek_xj:
	rjmp f_lseek_x
f_lseek:
;r17...r20		pointer from top of the file
	rcall f_check_open
	breq f_lseek_xj
	ldd r16,Y+fsize-fs
	cp r16,r17
	ldd r16,Y+fsize-fs+1
	cpc r16,r18
	ldd r16,Y+fsize-fs+2
	cpc r16,r19
	ldd r16,Y+fsize-fs+3
	cpc r16,r20
	brcc f_lseek1
	ldd r17,Y+fsize-fs
	ldd r18,Y+fsize-fs+1
	ldd r19,Y+fsize-fs+2
	ldd r20,Y+fsize-fs+3
f_lseek1:
	clr r16
	std Y+fptr-fs,r16
	std Y+fptr-fs+1,r16
	std Y+fptr-fs+2,r16
	std Y+fptr-fs+3,r16
f_lseek2:
	push r22
	push r23
	ldd r22,Y+csize-fs
	clr r23
	add r22,r22
	adc r23,r23
	ldd r16,Y+org_clust-fs
	std Y+curr_clust-fs,r16
	ldd r16,Y+org_clust-fs+1
	std Y+curr_clust-fs+1,r16
#ifdef _FS_FAT32
	ldd r16,Y+org_clust-fs+2
	std Y+curr_clust-fs+2,r16
	ldd r16,Y+org_clust-fs+3
	std Y+curr_clust-fs+3,r16
#endif
f_seek_lp:
	ldi r16,0
	cp r18,r22
	cpc r19,r23
	cpc r20,r16
	brcs f_lseek4
	push r17
	push r18
	push r19
	push r20
	ldd r17,Y+curr_clust-fs
	ldd r18,Y+curr_clust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+curr_clust-fs+2
	ldd r20,Y+curr_clust-fs+3
#endif
	rcall get_fat
	rcall check_clst_ll
	ldi r16,FR_DISK_ERR
	brcc f_lseek3
f_lseek_x1j:
	rjmp f_lseek_x1
f_lseek3:
	rcall check_clst_ul
	ldi r16,FR_DISK_ERR
	brcc f_lseek_x1j
	std Y+curr_clust-fs,r17
	std Y+curr_clust-fs+1,r18
#ifdef _FS_FAT32
	std Y+curr_clust-fs+2,r19
	std Y+curr_clust-fs+3,r20
#endif
	ldi r17,0
	ldd r16,Y+fptr-fs+1
	add r16,r22
	std Y+fptr-fs+1,r16
	ldd r16,Y+fptr-fs+2
	adc r16,r23
	std Y+fptr-fs+2,r16
	ldd r16,Y+fptr-fs+3
	adc r16,r17
	std Y+fptr-fs+3,r16
	pop r20
	pop r19
	pop r18
	pop r17
	ldi r16,0
	sub r18,r22
	sbc r19,r23
	sbc r20,r16
	rjmp f_seek_lp
f_lseek4:
	ldd r16,Y+fptr-fs
	add r16,r17
	std Y+fptr-fs,r16
	ldd r16,Y+fptr-fs+1
	adc r16,r18
	std Y+fptr-fs+1,r16
	ldd r16,Y+fptr-fs+2
	adc r16,r19
	std Y+fptr-fs+2,r16
	ldd r16,Y+fptr-fs+3
	adc r16,r20
	std Y+fptr-fs+3,r16
	ldd r17,Y+curr_clust-fs
	ldd r18,Y+curr_clust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+curr_clust-fs+2
	ldd r20,Y+curr_clust-fs+3
#endif
	rcall clust2sect
	ldi r16,FR_DISK_ERR
	breq f_lseek_x2
	ldd r22,Y+fptr-fs+1
	ldd r23,Y+fptr-fs+2
	lsr r23
	ror r22
	ldd r16,Y+csize-fs
	dec r16
	and r22,r16
	clr r16
	add r17,r22
	adc r18,r16
	adc r19,r16
	adc r20,r16
	std Y+dsect-fs,r17
	std Y+dsect-fs+1,r18
	std Y+dsect-fs+2,r19
	std Y+dsect-fs+3,r20
	ldi r16,FR_OK
	rjmp f_lseek_x2
f_lseek_x1:
	pop r20
	pop r19
	pop r18
	pop r17
f_lseek_x2:
	pop r23
	pop r22
f_lseek_x:
	ret

f_read_u_xj1:
	ldi r16,FR_WRONG_PARAM
f_read_u_xj:
	rjmp f_read_u_x1
f_read_u:
;Z:			ptr to buffer
;r18:r17:	number of bytes to read (must be multiple of 512)
	push r22
	push r23
	push r24
	push r25
	rcall f_check_open
	breq f_read_u_xj
	ldi r16,1
	and r16,r18
	or r16,r17
	brne f_read_u_xj1
	mov r22,r17
	mov r23,r18
	clr r24
	clr r25
	ldd r17,Y+fptr-fs
	ldd r18,Y+fptr-fs+1
	ldd r19,Y+fptr-fs+2
	ldd r20,Y+fptr-fs+3
	ldi r16,1
	and r16,r18
	or r16,r17
	brne f_read_u_xj1
	ldd XL,Y+csize-fs
	dec XL
	com XL
	mov r16,r23
	lsr r16
	dec r16
	and r16,XL
	brne f_read_u_lp
	ldi r16,0
	add r17,r22
	adc r18,r23
	adc r19,r16
	adc r20,r16
	brcc f_read_u2
f_read_u1:
	sub r17,r22
	sbc r18,r23
	sbci r19,0
	sbci r20,0
	rjmp f_read_u_lp
f_read_u2:
	rcall f_read_chk_fsize
	brcc f_read_u1
	mov r21,r19
	lsr r21
	mov r21,r18
	ror r21
	and r21,XL
	sub r17,r22
	sbc r18,r23
	sbci r19,0
	sbci r20,0
	mov r16,r19
	lsr r16
	mov r16,r18
	ror r16
	and r16,XL
	cp r16,r21
	brne f_read_u_lp
	mov r21,r23
	lsr r21
	rjmp f_read_u3
f_read_u_lp:
	subi r18,-2
	sbci r19,-1
	sbci r20,-1
	brcc f_read_u_t
	rcall f_read_chk_fsize
	brcc f_read_u_t
	subi r18,2
	sbci r19,0
	sbci r20,0
	ldi r21,1
f_read_u3:
	movw XH:XL,ZH:ZL
	push r21
	rcall f_read_sect
	pop r21
	tst r16
	brne f_read_u_x1
	add r21,r21
	clr r16
	add r18,r21
	adc r19,r16
	adc r20,r16
	add ZH,r21
	add r25,r21
	sub r23,r21
	brne f_read_u_lp
	ldi r16,FR_OK
	std Y+fptr-fs,r17
	std Y+fptr-fs+1,r18
	std Y+fptr-fs+2,r19
	std Y+fptr-fs+3,r20
	mov r17,r24
	mov r18,r25
f_read_u_x1:
	pop r25
	pop r24
	pop r23
	pop r22
	ret

f_read_u_t:
	subi r18,2
	sbci r19,0
	sbci r20,0
	rjmp f_read1

f_read:
;Z:			ptr to buffer
;r18:r17:	number of bytes to read
	rcall f_check_open
	breq f_read_x
	push r22
	push r23
	push r24
	push r25
	mov r22,r17
	mov r23,r18
	clr r24
	clr r25
	ldd r17,Y+fptr-fs
	ldd r18,Y+fptr-fs+1
	ldd r19,Y+fptr-fs+2
	ldd r20,Y+fptr-fs+3
f_read1:
	mov r16,r18
	andi r16,HIGH(511)
	or r16,r17
	breq f_read_lp
	ldi XL,LOW(sector)
	ldi XH,HIGH(sector)
	ldi r21,1
	rcall f_read_sect
	tst r16
	brne f_read_x1
	mov XL,r17
	mov XH,r18
	andi XH,HIGH(511)
	subi XL,LOW(-sector)
	sbci XH,HIGH(-sector)
f_read_lp:
	rcall f_read_chk_fsize
	ldi r16,FR_END_OF_FILE
	brcc f_read_x2
	mov r16,r18
	andi r16,HIGH(511)
	or r16,r17
	brne f_read2
	ldi XL,LOW(sector)
	ldi XH,HIGH(sector)
	ldi r21,1
	rcall f_read_sect
	tst r16
	brne f_read_x1
	ldi XL,LOW(sector)
	ldi XH,HIGH(sector)
f_read2:
	ld r16,X+
	st Z+,r16
	subi r17,-1
	sbci r18,-1
	sbci r19,-1
	sbci r20,-1
	adiw r25:r24,1
	subi r22,1
	sbci r23,0
	brne f_read_lp
	ldi r16,FR_OK
f_read_x2:
	std Y+fptr-fs,r17
	std Y+fptr-fs+1,r18
	std Y+fptr-fs+2,r19
	std Y+fptr-fs+3,r20
	mov r17,r24
	mov r18,r25
f_read_x1:
	pop r25
	pop r24
	pop r23
	pop r22
f_read_x:
	ret

f_read_chk_fsize:
	ldd r16,Y+fsize-fs
	cp r17,r16
	ldd r16,Y+fsize-fs+1
	cpc r18,r16
	ldd r16,Y+fsize-fs+2
	cpc r19,r16
	ldd r16,Y+fsize-fs+3
	cpc r20,r16
	ret

f_check_open:
	rcall check_mount
	breq f_check_open_x
	ldd r16,Y+flag-fs
	andi r16,FA_OPENED
	ldi r16,FR_NOT_OPENED
f_check_open_x:
	ret

f_read_sect:
	push r17
	push r18
	push r19
	push r20
	push r21
	mov r16,r18
	lsr r16
	ldd r21,Y+csize-fs
	dec r21
	and r21,r16
	breq f_read_sect1
	ldd r17,Y+dsect-fs
	ldd r18,Y+dsect-fs+1
	ldd r19,Y+dsect-fs+2
	ldd r20,Y+dsect-fs+3
	rjmp f_read_sect4
f_read_sect1:
	push XL
	push XH
	ldi r16,-1
	eor r16,r21
	add r16,r16
	and r16,r18
	or r16,r19
	or r16,r20
	brne f_read_sect2
	ldd r17,Y+org_clust-fs
	ldd r18,Y+org_clust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+org_clust-fs+2
	ldd r20,Y+org_clust-fs+3
#endif
	rjmp f_read_sect3
f_read_sect2:
	ldd r17,Y+curr_clust-fs
	ldd r18,Y+curr_clust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+curr_clust-fs+2
	ldd r20,Y+curr_clust-fs+3
#endif
	push r21
	rcall get_fat
	pop r21
f_read_sect3:
	pop XH
	pop XL
	rcall check_clst_ll
	brcs f_read_sect_x11
	std Y+curr_clust-fs,r17
	std Y+curr_clust-fs+1,r18
#ifdef _FS_FAT32
	std Y+curr_clust-fs+2,r19
	std Y+curr_clust-fs+3,r20
#endif
	rcall clust2sect
	breq f_read_sect_x11
	ldi r16,0
	add r17,r21
	adc r18,r16
	adc r19,r16
	adc r20,r16
f_read_sect4:
	pop r16
	push r17
	push r18
	push r19
	push r20
	push r16
	rcall disk_read
	pop r21
	pop r20
	pop r19
	pop r18
	pop r17
	tst r16
	brne f_read_sect_x1
	add r17,r21
	adc r18,r16
	adc r19,r16
	adc r20,r16
	std Y+dsect-fs,r17
	std Y+dsect-fs+1,r18
	std Y+dsect-fs+2,r19
	std Y+dsect-fs+3,r20
f_read_sect_x2:
	pop r20
	pop r19
	pop r18
	pop r17
	ret
f_read_sect_x11:
	pop r16
f_read_sect_x1:
	ldi r16,0
	std Y+flag-fs,r16
	ldi r16,FR_DISK_ERR
	rjmp f_read_sect_x2

f_open:
;Z:				ptr to path name
;pname_in_ram:	0 - path in flash
;				1 - path in RAM
	rcall check_mount
	breq f_open_x
	clr r16
	std Y+flag-fs,r16
	in r17,status
	in XL,sp_low
	in XH,sp_high
	sbiw X,12
	cli
	out sp_low,XL
	out sp_high,XH
	out status,r17
	adiw X,1
	std Y+fn-fs,XL
	std Y+fn-fs+1,XH
	rcall follow_path
	tst r16
	brne f_open_x1
	ldi r16,FR_NO_FILE
	ldd r17,Z+0
	tst r17
	breq f_open_x1
	ldd r17,Z+DIR_Attr
	andi r17,AM_DIR
	brne f_open_x1
	ldi r16,FA_OPENED
	sbrc r17,AM_RDO_bp
	ori r16,FA_WPRT
	std Y+flag-fs,r16
	rcall get_clust
	std Y+org_clust-fs,r17
	std Y+org_clust-fs+1,r18
	std Y+org_clust-fs+2,r19
	std Y+org_clust-fs+3,r20
	ldd r17,Z+DIR_FileSize
	ldd r18,Z+DIR_FileSize+1
	ldd r19,Z+DIR_FileSize+2
	ldd r20,Z+DIR_FileSize+3
	std Y+fsize-fs,r17
	std Y+fsize-fs+1,r18
	std Y+fsize-fs+2,r19
	std Y+fsize-fs+3,r20
	clr r16
	std Y+fptr-fs,r16
	std Y+fptr-fs+1,r16
	std Y+fptr-fs+2,r16
	std Y+fptr-fs+3,r16
	ldi r16,FR_OK
f_open_x1:
	in r17,status
	in XL,sp_low
	in XH,sp_high
	adiw X,12
	cli
	out sp_low,XL
	out sp_high,XH
	out status,r17
f_open_x:
	ret

follow_path:
;Z:				ptr to path name
;pname_in_ram:	0 - path in flash
;				1 - path in RAM
follow_path_lp1:
	rcall get_path_chr
	cpi r16,' '
	brne follow_path1
	adiw Z,1
	rjmp follow_path_lp1
follow_path1:
	cpi r16,'/'
	brne follow_path2
	adiw Z,1
follow_path2:
	clr r16
	std Y+sclust-fs,r16
	std Y+sclust-fs+1,r16
#ifdef _FS_FAT32
	std Y+sclust-fs+2,r16
	std Y+sclust-fs+3,r16
#endif
	rcall get_path_chr
	cpi r16,' '
	brcc follow_path_lp2
	rcall dir_rewind
	clr r17
	sts sector,r17
	rjmp follow_path_x
follow_path_lp2:
	rcall create_name
	push ZL
	push ZH
	rcall dir_find
	tst r16
	brne follow_path_x1
	ldd XL,Y+fn-fs
	ldd XH,Y+fn-fs+1
	adiw X,11
	ld r17,X
	tst r17
	brne follow_path_x1
	ldd r16,Z+DIR_Attr
	andi r16,AM_DIR
	ldi r16,FR_NO_FILE
	breq follow_path_x1
	rcall get_clust
	std Y+sclust-fs,r17
	std Y+sclust-fs+1,r18
	std Y+sclust-fs+2,r19
	std Y+sclust-fs+3,r20
	pop ZH
	pop ZL
	rjmp follow_path_lp2
follow_path_x1:
	pop r17
	pop r17
follow_path_x:
	ret

dir_next:
;Y:			ptr to fs
	ldd XL,Y+index-fs
	ldd XH,Y+index-fs+1
	adiw X,16
	brne dir_next1
	ldi r16,FR_NO_FILE
	rjmp dir_next_x
dir_next1:
	std Y+index-fs,XL
	std Y+index-fs+1,XH
	subi r17,-1
	sbci r18,-1
	sbci r19,-1
	sbci r20,-1
	ldd r16,Y+clust-fs
	ldd r21,Y+clust-fs+1
	or r16,r21
#ifdef _FS_FAT32
	ldd r21,Y+clust-fs+2
	or r16,r21
	ldd r21,Y+clust-fs+3
	or r16,r21
#endif
	brne dir_next2
	ldd r16,Y+n_rootdir-fs
	cp XL,r16
	ldd r16,Y+n_rootdir-fs+1
	cpc XH,r16
	brcs dir_next_x1
	ldi r16,FR_NO_FILE
	rjmp dir_next_x
dir_next2:
	lsr XH
	ror XL
	lsr XH
	ror XL
	lsr XH
	ror XL
	lsr XH
	ror XL
	ldd r16,Y+csize-fs
	dec r16
	and XL,r16
	brne dir_next_x1
	ldd r17,Y+clust-fs
	ldd r18,Y+clust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+clust-fs+2
	ldd r20,Y+clust-fs+3
#endif
	rcall get_fat
	rcall check_clst_ll
	brcc dir_next3
	ldi r16,FR_DISK_ERR
	rjmp dir_next_x
dir_next3:
	rcall check_clst_ul
	brcs dir_next4
	ldi r16,FR_NO_FILE
	rjmp dir_next_x
dir_next4:
	std Y+clust-fs,r17
	std Y+clust-fs+1,r18
	std Y+clust-fs+2,r19
	std Y+clust-fs+3,r20
	rcall clust2sect
dir_next_x1:
	ldi r16,FR_OK
dir_next_x:
	ret

dir_find:
;Y:			ptr to fs
	rcall dir_rewind
	cpi r16,FR_OK
	brne dir_find_x
	ldd r17,Y+sect-fs
	ldd r18,Y+sect-fs+1
	ldd r19,Y+sect-fs+2
	ldd r20,Y+sect-fs+3
dir_find_lp1:
	push r17
	push r18
	push r19
	push r20
	ldi XL,LOW(sector)
	ldi XH,HIGH(sector)
	movw ZH:ZL,XH:XL
	rcall disk_read_default
	cpi r16,FR_OK
	breq dir_find1
	ldi r16,FR_DISK_ERR
	rjmp dir_find_x1
dir_find1:
	ldi r19,16
	ldd XL,Y+fn-fs
	ldd XH,Y+fn-fs+1
	adiw X,11
	ld r20,X
dir_find_lp2:
	ldd r16,Z+DIR_Name
	tst r16
	ldi r16,FR_NO_FILE
	breq dir_find_x1
	push ZL
	push ZH
	ldd r16,Z+DIR_Attr
	cpi r16,AM_LFN
	sbrc r20,7			;skip if normal file search, only to include deleted LFN slots for file creation
	breq dir_find2
	andi r16,AM_VOL
	brne dir_find3
dir_find2:
	sbrs r20,7
	rcall dir_find_cmp
	sbrc r20,7
	rcall dir_find_empty
	brne dir_find3
	pop ZH
	pop ZL
	std Y+index-fs,ZL
	std Y+index-fs+1,ZH
	ldi r16,FR_OK
	rjmp dir_find_x1
dir_find3:
	pop ZH
	pop ZL
	adiw Z,32
	dec r19
	brne dir_find_lp2
	pop r20
	pop r19
	pop r18
	pop r17
	rcall dir_next
	cpi r16,FR_OK
	breq dir_find_lp1
	rjmp dir_find_x
dir_find_x1:
	pop r17
	pop r17
	pop r17
	pop r17
dir_find_x:
	ret

dir_find_cmp:
	ldd XL,Y+fn-fs
	ldd XH,Y+fn-fs+1
	ldi r18,11
dir_find_lp3:
	ld r16,X+
	ld r17,Z+
	cp r16,r17
	brne dir_find_cmpx
	dec r18
	brne dir_find_lp3
dir_find_cmpx:
	ret

dir_find_empty:
	ld r16,Z
	tst r16
	breq dir_find_emptyx
	cpi r16,$e5
dir_find_emptyx:
	ret

get_path_chr:
	sbis pname_in_ram
	lpm r16,Z
	sbic pname_in_ram
	ld r16,Z
	ret

create_name:
;Y:			ptr to fs
	ldd XL,Y+fn-fs
	ldd XH,Y+fn-fs+1
	movw r19:r18,XH:XL
	ldi r16,' '
	ldi r17,11
	rcall mem_set
	movw XH:XL,r19:r18
	clr r17
	ldi r20,8
create_name_lp:
	rcall get_path_chr
	adiw Z,1
	cpi r16,' '
	brcs create_name_x
	cpi r16,'/'
	breq create_name_x
	cpi r16,'.'
	breq create_name1
	cp r17,r20
	brne create_name2
create_name1:
	cpi r16,'.'
	brne create_name_x
	cpi r20,8
	brne create_name_x
	movw XH:XL,r19:r18
	adiw X,8
	ldi r17,8
	ldi r20,11
	rjmp create_name_lp
create_name2:
	st X+,r16
	inc r17
	rjmp create_name_lp
create_name_x:
	cpi r16,' '
	ldi r16,1
	brcs create_name_x1
	ldi r16,0
create_name_x1:
	movw XH:XL,r19:r18
	adiw X,11
	st X,r16
	ret

mem_set:
	st X+,r16
	dec r17
	brne mem_set
	ret

dir_rewind:
;Y:			ptr to fs
	clr r16
	std Y+index-fs,r16
	std Y+index-fs+1,r16
	ldd r17,Y+sclust-fs
	ldd r18,Y+sclust-fs+1
#ifdef _FS_FAT32
	ldd r19,Y+sclust-fs+2
	ldd r20,Y+sclust-fs+3
#endif
	clr r16
	cpi r17,1
	cpc r18,r16
#ifdef _FS_FAT32
	cpc r19,r16
	cpc r20,r16
#endif
	brne dir_rewind2
dir_rewind1:
	ldi r16,FR_DISK_ERR
	rjmp dir_rewind_x
dir_rewind2:
	rcall check_clst_ul
	brcc dir_rewind1
#ifdef _FS_FAT32
	ldd r16,Y+fs_type-fs
	cpi r16,FS_FAT32
	brne dir_rewind3
	mov r16,r17
	or r16,r18
	or r16,r19
	or r16,r20
	brne dir_rewind3
	ldd r17,Y+dirbase-fs
	ldd r18,Y+dirbase-fs+1
	ldd r19,Y+dirbase-fs+2
	ldd r20,Y+dirbase-fs+3
dir_rewind3:
#endif
	std Y+clust-fs,r17
	std Y+clust-fs+1,r18
#ifdef _FS_FAT32
	std Y+clust-fs+2,r19
	std Y+clust-fs+3,r20
#endif
	mov r16,r17
	or r16,r18
#ifdef _FS_FAT32
	or r16,r19
	or r16,r20
#endif
	brne dir_rewind4
	ldd r17,Y+dirbase-fs
	ldd r18,Y+dirbase-fs+1
	ldd r19,Y+dirbase-fs+2
	ldd r20,Y+dirbase-fs+3
	rjmp dir_rewind5
dir_rewind4:
	rcall clust2sect
dir_rewind5:
	std Y+sect-fs,r17
	std Y+sect-fs+1,r18
	std Y+sect-fs+2,r19
	std Y+sect-fs+3,r20
	ldi r16,FR_OK
dir_rewind_x:
	ret

check_clst_ll:
	clr r16
	cpi r17,2
	cpc r18,r16
#ifdef _FS_FAT32
	cpc r19,r16
	cpc r20,r16
#endif
	ret

check_clst_ul:
	ldd r16,Y+n_fatent-fs
	cp r17,r16
	ldd r16,Y+n_fatent-fs+1
	cpc r18,r16
#ifdef _FS_FAT32
	ldd r16,Y+n_fatent-fs+2
	cpc r19,r16
	ldd r16,Y+n_fatent-fs+3
	cpc r20,r16
#endif
	ret

get_fat:
	push ZL
	push ZH
	ldi ZL,LOW(sector)
	ldi ZH,HIGH(sector)
	rcall get_fat_
	pop ZH
	pop ZL
	ret

get_fat_:
;Y:				ptr to fs
;r17...r20		cluster number
	push r22
	push r23
	rcall check_clst_ll
	brcc get_fat1
get_fat_x1j:
	rjmp get_fat_x1
get_fat1:
	rcall check_clst_ul
	brcc get_fat_x1j
	ldd r16,Y+fs_type-fs
#ifdef _FS_FAT12
	cpi r16,FS_FAT12
	breq get_fat2
	rjmp get_fat5
get_fat2:
	mov r21,r17
	mov XL,r17
	mov XH,r18
	mov r19,XL
	mov r20,XH
	lsr r20
	ror r19
	add XL,r19
	adc XH,r20
	mov r18,XH
	andi XH,1
	add XL,ZL
	adc XH,ZH
	push XL
	push XH
	lsr r18
	ldi r16,0
	ldd r17,Y+fatbase-fs
	add r17,r18
	ldd r18,Y+fatbase-fs+1
	adc r18,r16
	ldd r19,Y+fatbase-fs+2
	adc r19,r16
	ldd r20,Y+fatbase-fs+3
	push r17
	push r18
	push r19
	push r20
	ldi r16,1
	movw X,Z
	rcall disk_read
	pop r20
	pop r19
	pop r18
	pop r17
	pop XH
	pop XL
	tst r16
get_fat_x1j1:
	brne get_fat_x1j
	ld r22,X+
	ld r23,X
	subi XH,HIGH(512)
	cp XL,ZL
	cpc XH,ZH
	brne get_fat3
	subi r17,-1
	sbci r18,-1
	sbci r19,-1
	sbci r20,-1
	ldi r16,1
	movw X,Z
	rcall disk_read
	tst r16
	brne get_fat_x1j1
	ld r23,Z
get_fat3:
	mov r17,r22
	mov r18,r23
	andi r21,1
	breq get_fat4
	lsr r18
	ror r17
	lsr r18
	ror r17
	lsr r18
	ror r17
	lsr r18
	ror r17
get_fat4:
	andi r18,$f
	rjmp get_fat_x
get_fat5:
#endif
#ifdef _FS_FAT16
	cpi r16,FS_FAT16
	brne get_fat6
	push r17
	push r18
	ldi r16,0
	ldd r17,Y+fatbase-fs
	add r17,r18
	ldd r18,Y+fatbase-fs+1
	adc r18,r16
	ldd r19,Y+fatbase-fs+2
	adc r19,r16
	ldd r20,Y+fatbase-fs+3
	adc r20,r16
	ldi r16,1
	movw X,Z
	rcall disk_read
	pop XH
	pop XL
	tst r16
	brne get_fat_x1
	andi XL,$fe
	andi XH,1
	add XL,ZL
	adc XH,ZH
	ld r17,X+
	ld r18,X
	rjmp get_fat_x
get_fat6:
#endif
#ifdef _FS_FAT32
	cpi r16,FS_FAT32
	brne get_fat7
	push r17
	add r17,r17
	adc r18,r18
	adc r19,r19
	adc r20,r20
	ldd r17,Y+fatbase-fs
	add r17,r18
	ldd r18,Y+fatbase-fs+1
	adc r18,r19
	ldd r19,Y+fatbase-fs+2
	adc r19,r20
	ldd r20,Y+fatbase-fs+3
	ldi r16,0
	adc r20,r16
	ldi r16,1
	movw X,Z
	rcall disk_read
	pop XL
	tst r16
	brne get_fat_x1
	andi XL,$7f
	clr XH
	add XL,XL
	adc XH,XH
	add XL,XL
	adc XH,XH
	add XL,ZL
	adc XH,ZH
	ld r17,X+
	ld r18,X+
	ld r19,X+
	ld r20,X
	andi r20,$0f
	rjmp get_fat_x
get_fat7:
#endif
get_fat_x1:
	ldi r16,1
get_fat_x:
	pop r23
	pop r22
	ret

clust2sect:
;Y:				ptr to fs
;r17...r20		cluster number
	rcall check_clst_ul
	brcs clust2sect1
	clr r17
	clr r18
	clr r19
	clr r20
	rjmp clust2sect_x
clust2sect1:
	subi r17,2
	sbci r18,0
#ifdef _FS_FAT32
	sbci r19,0
	sbci r20,0
#endif
#ifndef _FS_FAT32
	clr r19
	clr r20
#endif
	ldd r16,Y+csize-fs
clust2sect1_lp:
	lsr r16
	brcs clust2sect2
	add r17,r17
	adc r18,r18
	adc r19,r19
	adc r20,r20
	rjmp clust2sect1_lp
clust2sect2:
	ldd r16,Y+database-fs
	add r17,r16
	ldd r16,Y+database-fs+1
	adc r18,r16
	ldd r16,Y+database-fs+2
	adc r19,r16
	ldd r16,Y+database-fs+3
	adc r20,r16
	mov r16,r17
	or r16,r18
	or r16,r19
	or r16,r20
clust2sect_x:
	ret

get_clust:
;Y:			ptr to fs
;Z			ptr to dir entry
	clr r19
	clr r20
	ldd r17,Z+DIR_FstClusLO
	ldd r18,Z+DIR_FstClusLO+1
#if _FS_FAT32
	ldd r16,Y+fs_type-fs
	cpi r16,FS_FAT32
	brne get_clust_x
	ldd r19,Z+DIR_FstClusHI
	ldd r20,Z+DIR_FstClusHI+1
#endif
get_clust_x:
	ret

check_mount:
	ldi YL,LOW(fs)
	ldi YH,HIGH(fs)
	ldd r16,Y+fs_type-fs
	tst r16
	ldi r16,FR_NOT_ENABLED
	ret

mount:
	rcall disk_initialize
	andi r16,STA_NOINIT
	ldi r16,FR_NOT_READY
	breq mount1
	rjmp mount_x
mount1:
	ldi YL,LOW(fs)
	ldi YH,HIGH(fs)
	clr r17
	clr r18
	clr r19
	clr r20
	std Y+fatbase-fs,r17
	std Y+fatbase-fs+1,r18
	std Y+fatbase-fs+2,r19
	std Y+fatbase-fs+3,r20
	rcall check_fs
	cpi r16,1
	brne mount2
	ldi ZL,LOW(sector+MBR_Table)
	ldi ZH,HIGH(sector+MBR_Table)
	ldd r17,Z+4
	tst r17
	breq mount2
	ldd r17,Z+8
	ldd r18,Z+9
	ldd r19,Z+10
	ldd r20,Z+11
	std Y+fatbase-fs,r17
	std Y+fatbase-fs+1,r18
	std Y+fatbase-fs+2,r19
	std Y+fatbase-fs+3,r20
	rcall check_fs
mount2:
	cpi r16,3
	brne mount3
	ldi r16,FR_DISK_ERR
	rjmp mount_x
mount3:
	tst r16
	ldi r16,FR_NO_FILESYSTEM
	breq mount4
	rjmp mount_x
mount4:
	ldi ZL,LOW(sector)
	ldi ZH,HIGH(sector)
	ldd r16,Z+BPB_FATSz16
	ldd r17,Z+BPB_FATSz16+1
	clr r18
	clr r19
	mov r20,r16
	or r20,r17
	brne mount5
	ldd r16,Z+BPB_FATSz32
	ldd r17,Z+BPB_FATSz32+1
	ldd r18,Z+BPB_FATSz32+2
	ldd r19,Z+BPB_FATSz32+3
mount5:
	std Y+fatsize-fs,r16
	std Y+fatsize-fs+1,r17
	std Y+fatsize-fs+2,r18
	std Y+fatsize-fs+3,r19
	ldd r20,Z+BPB_NumFATs
mount5_lp:
	dec r20
	breq mount6
	add r16,r16
	adc r17,r17
	adc r18,r18
	adc r19,r19
	rjmp mount5_lp
mount6:
	std Y+fsize-fs,r16
	std Y+fsize-fs+1,r17
	std Y+fsize-fs+2,r18
	std Y+fsize-fs+3,r19
	ldd r20,Z+BPB_RsvdSecCnt
	ldd r21,Z+BPB_RsvdSecCnt+1
	ldd r16,Y+fatbase-fs
	ldd r17,Y+fatbase-fs+1
	ldd r18,Y+fatbase-fs+2
	ldd r19,Y+fatbase-fs+3
	add r16,r20
	adc r17,r21
	ldi r20,0
	adc r18,r20
	adc r19,r20
	std Y+fatbase-fs,r16
	std Y+fatbase-fs+1,r17
	std Y+fatbase-fs+2,r18
	std Y+fatbase-fs+3,r19
	ldd r16,Z+BPB_SecPerClus
	std Y+csize-fs,r16
	ldd r16,Z+BPB_RootEntCnt
	ldd r17,Z+BPB_RootEntCnt+1
	std Y+n_rootdir-fs,r16
	std Y+n_rootdir-fs+1,r17
	ldd r16,Z+BPB_TotSec16
	ldd r17,Z+BPB_TotSec16+1
	clr r18
	clr r19
	mov r20,r16
	or r20,r17
	brne mount7
	ldd r16,Z+BPB_TotSec32
	ldd r17,Z+BPB_TotSec32+1
	ldd r18,Z+BPB_TotSec32+2
	ldd r19,Z+BPB_TotSec32+3
mount7:
	ldd r20,Z+BPB_RsvdSecCnt
	ldd r21,Z+BPB_RsvdSecCnt+1
	sub r16,r20
	sbc r17,r21
	ldi r20,0
	sbc r18,r20
	sbc r19,r20
	ldd r20,Y+fsize-fs
	sub r16,r20
	ldd r20,Y+fsize-fs+1
	sbc r17,r20
	ldd r20,Y+fsize-fs+2
	sbc r18,r20
	ldd r20,Y+fsize-fs+3
	sbc r19,r20
	ldd r20,Y+n_rootdir-fs
	ldd r21,Y+n_rootdir-fs+1
	lsr r21
	ror r20
	lsr r21
	ror r20
	lsr r21
	ror r20
	lsr r21
	ror r20
	push r21
	push r20
	sub r16,r20
	sbc r17,r21
	ldi r20,0
	sbc r18,r20
	sbc r19,r20
	ldd r20,Y+csize-fs
mount7_lp:
	lsr r20
	brcs mount8
	lsr r19
	ror r18
	ror r17
	ror r16
	rjmp mount7_lp
mount8:
	subi r16,LOW(-2)
	sbci r17,HIGH(-2)
	sbci r18,BYTE3(-2)
	sbci r19,BYTE4(-2)
	std Y+n_fatent-fs,r16
	std Y+n_fatent-fs+1,r17
#ifdef _FS_FAT32
	std Y+n_fatent-fs+2,r18
	std Y+n_fatent-fs+3,r19
#endif
	ldi r21,0
	cpi r16,LOW($ff7)
	ldi r20,HIGH($ff7)
	cpc r17,r20
	ldi r20,0
	cpc r18,r20
	cpc r19,r20
#ifdef _FS_FAT12
	brcc mount9
	ldi r21,FS_FAT12
	rjmp mount11
mount9:
#else
	brcs mount11
#endif
	cpi r16,LOW($fff7)
	ldi r20,HIGH($fff7)
	cpc r17,r20
	ldi r20,0
	cpc r18,r20
	cpc r19,r20
#ifdef _FS_FAT16
	brcc mount10
	ldi r21,FS_FAT16
	rjmp mount11
mount10:
#else
	brcs mount11
#endif
#if _FS_FAT32
	ldi r21,FS_FAT32
#endif
mount11:
	std Y+fs_type-fs,r21
#ifdef _FS_FAT32
	cpi r21,FS_FAT32
	brne mount12
	ldd r16,Z+BPB_RootClus
	ldd r17,Z+BPB_RootClus+1
	ldd r18,Z+BPB_RootClus+2
	ldd r19,Z+BPB_RootClus+3
	rjmp mount13
mount12:
#endif
	rcall mount_calcbase
mount13:
	std Y+dirbase-fs,r16
	std Y+dirbase-fs+1,r17
	std Y+dirbase-fs+2,r18
	std Y+dirbase-fs+3,r19
#ifdef _FS_FAT32
	cpi r21,FS_FAT32
	brne mount14
	rcall mount_calcbase
mount14:
#endif
	pop r20
	add r16,r20
	pop r20
	adc r17,r20
	ldi r20,0
	adc r18,r20
	adc r19,r20
	std Y+database-fs,r16
	std Y+database-fs+1,r17
	std Y+database-fs+2,r18
	std Y+database-fs+3,r19
	std Y+flag-fs,r20
	ldi r16,FR_OK
mount_x:
	ret

mount_calcbase:
	ldd r16,Y+fatbase-fs
	ldd r17,Y+fatbase-fs+1
	ldd r18,Y+fatbase-fs+2
	ldd r19,Y+fatbase-fs+3
	ldd r20,Y+fsize-fs
	add r16,r20
	ldd r20,Y+fsize-fs+1
	adc r17,r20
	ldd r20,Y+fsize-fs+2
	adc r18,r20
	ldd r20,Y+fsize-fs+3
	adc r19,r20
	ret

check_fs:
	rcall disk_read_default
	tst r16
	ldi r16,3
	brne check_fs_x
	ldi r16,2
	lds r17,(sector+BS_55AA)
	cpi r17,$55
	brne check_fs_x
	lds r17,(sector+BS_55AA+1)
	cpi r17,$aa
	brne check_fs_x
	ldi r16,0
	lds r17,(sector+BS_FilSysType)
	cpi r17,'F'
	brne check_fs1
	lds r17,(sector+BS_FilSysType+1)
	cpi r17,'A'
	breq check_fs_x
check_fs1:
#ifdef _FS_FAT32
	lds r17,(sector+BS_FilSysType32)
	cpi r17,'F'
	brne check_fs2
	lds r17,(sector+BS_FilSysType32+1)
	cpi r17,'A'
	breq check_fs_x
check_fs2:
#endif
	ldi r16,1
check_fs_x:
	ret

dly_1000us:
	ldi r16,9
dly_1000us_lp:
	rcall dly_100us
	dec r16
	brne dly_1000us_lp
dly_100us:
	push r24
	push r25
	ldi r24,LOW(int(100/(1.0/clock*4)))
	ldi r25,HIGH(int(100/(1.0/clock*4)))
dly_us_lp:
	sbiw r25:r24,1	;2
	brne dly_us_lp	;2
	pop r25
	pop r24
	ret

rcvr_mmc:
;X: 		ptr to buffer
;r19:r18:	byte count
rcvr_mmc_lp:
	call read_mmc
	st X+,r16
rcrv_mmc1:
	subi r18,LOW(1)
	sbci r19,HIGH(1)
	brne rcvr_mmc_lp
	ret

wait_ready:
	push r24
	push r25
	ldi r24,LOW(5000)
	ldi r25,HIGH(5000)
wait_ready_lp:
	call read_mmc
	cpi r16,-1
	breq wait_ready_x
	rcall dly_100us
	sbiw r25:r24,1
	brne wait_ready_lp
wait_ready_x:
	ldi r16,1
	or r24,r25
	brne wait_ready_x1
	clr r16
wait_ready_x1:
	pop r25
	pop r24
	ret		

select:
	cbi CS
	call read_mmc_1st
	rcall wait_ready
	cpi r16,1
	breq select_x
	call deselect
	clr r16
select_x:
	ret

rcvr_datablock:
;X: 		ptr to buffer
;r19:r18:	byte count
	push r24
	push r25
	ldi r24,LOW(1000)
	ldi r25,HIGH(1000)
rcvr_datablock_lp:
	call read_mmc
	cpi r16,$fe
	breq rcvr_datablock2
	cpi r16,-1
	brne rcvr_datablock1
	rcall dly_100us
	sbiw r25:r24,1
	brne rcvr_datablock_lp
rcvr_datablock1:
	clr r16
	rjmp rcvr_datablock_x
rcvr_datablock2:
	rcall rcvr_mmc
	call read_mmc_1st		;discard 2 bytes CRC
	call read_mmc
	ldi r16,1
rcvr_datablock_x:
	pop r25
	pop r24
	ret

send_cmd_p0:
	clr r17
	clr r18
	clr r19
	clr r20
send_cmd:
;r16:		command byte
;r17-r20	Argument
	push r17
	push r18
	push r19
	push r20
	tst r16
	brpl send_cmd1
	andi r16,$7f
	push r16
	ldi r16,CMD55
	rcall send_cmd_p0
	cpi r16,2
	brcc send_cmd_x1
	pop r16
send_cmd1:
	push r16
	cpi r16,CMD12
	breq send_cmd2
	call deselect
	rcall select
	tst r16
	brne send_cmd2
	ldi r16,-1
	rjmp send_cmd_x1
send_cmd2:
	pop r16
	mov r17,r16
	ori r16,$40
	call send_mmc
	pop r16
	call send_mmc
	pop r16
	call send_mmc
	pop r16
	call send_mmc
	pop r16
	call send_mmc
	ldi r16,1
	cpi r17,CMD0
	brne send_cmd3
	ldi r16,$95
send_cmd3:
	cpi r17,CMD8
	brne send_cmd4
	ldi r16,$87
send_cmd4:
	call send_mmc
	cpi r17,CMD12
	brne send_cmd5
	call read_mmc_1st
send_cmd5:
	ldi r17,10
send_cmd_lp:
	call read_mmc
	tst r16
	brpl send_cmd_x
	dec r17
	brne send_cmd_lp
	rjmp send_cmd_x
send_cmd_x1:
	pop r17
	pop r17
	pop r17
	pop r17
	pop r17
send_cmd_x:
	ret

disk_status:
	lds r16,(dstat)
	sbrc r16,STA_NOINIT_bp
	rjmp disk_status_x
	ldi r16,CMD13
	rcall send_cmd_p0
	tst r16
	breq disk_status1
	ldi r16,STA_NOINIT
	sts (dstat),r16
disk_status1:
	call read_mmc_1st
	call deselect
	lds r16,(dstat)
disk_status_x:
	ret

disk_initialize:
	cbi SD_PWR
	call delay2ms
#ifdef cache_sector
	cbi valid_sect
#endif
	ldi r17,10
disk_init_lp1:
	call read_mmc
	dec r17
	brne disk_init_lp1
	ldi r21,0
	ldi r16,CMD0
	rcall send_cmd_p0
	cpi r16,1
	brne disk_init_xj
	ldi r16,CMD8
	ldi r17,LOW($1aa)
	ldi r18,HIGH($1aa)
	clr r19
	clr r29
	rcall send_cmd
	cpi r16,1
	brne disk_init10
	call read_mmc_1st
	call read_mmc
	call read_mmc
	mov r17,r16
	call read_mmc
	ldi r24,LOW(1000)
	ldi r25,HIGH(1000)
	cpi r17,1
	brne disk_init1
	cpi r16,$aa
	brne disk_init1
disk_init_lp2:
	ldi r16,ACMD41
	ldi r17,LOW(1<<30)
	ldi r18,HIGH(1<<30)
	ldi r19,BYTE3(1<<30)
	ldi r20,BYTE4(1<<30)
	rcall send_cmd
	tst r16
	breq disk_init1
	rcall dly_1000us
	sbiw r25:r24,1
	brne disk_init_lp2
disk_init_xj:
	rjmp disk_init_x
disk_init1:
	ldi r16,CMD58
	rcall send_cmd_p0
	tst r16
	brne disk_init_x
	call read_mmc_1st
	mov r17,r16
	call read_mmc
	call read_mmc
	call read_mmc
	ldi r21,CT_SD2
	sbrc r17,6
	ori r21,CT_BLOCK
	rjmp disk_init_x
disk_init10:
#ifdef legacy_support
	ldi r16,ACMD41
	rcall send_cmd_p0
	ldi r22,ACMD41
	ldi r21,CT_SD1
	cpi r16,2
	brcs disk_init11
	ldi r22,CMD1
	ldi r21,CT_MMC
disk_init11:
	ldi r24,LOW(1000)
	ldi r25,HIGH(1000)
disk_init_lp3:
	mov r16,r22
	rcall send_cmd_p0
	tst r16
	breq disk_init12
	rcall dly_1000us
	sbiw r25:r24,1
	brne disk_init_lp3
	clr r21
	rjmp disk_init_x
disk_init12:
	ldi r16,CMD16
	ldi r17,LOW(512)
	ldi r18,HIGH(512)
	clr r19
	clr r20
	rcall send_cmd
	tst r16
	breq disk_init_x
#endif
	clr r21
disk_init_x:
	sts (cardtype),r21
	ldi r16,0
	tst r21
	brne disk_init_x1
	ldi r16,STA_NOINIT
disk_init_x1:
	sts (dstat),r16
	push r16
	call deselect
	pop r16
	ret

disk_read_default:
	ldi XL,LOW(sector)
	ldi XH,HIGH(sector)
	ldi r16,1
disk_read:
;X:			ptr to buffer
;Y			ptr to fs
;r16		sector count (1...128)
;r17...r20	start sector number
#ifdef cache_sector
	cbi cache_flag
	cpi r16,1
	brne disk_read0
	ldi r16,HIGH(sector)
	cpi XL,LOW(sector)
	cpc XH,r16
	brne disk_read01
	sbis valid_sect
	rjmp disk_read02
	ldd r16,Y+last_sect-fs
	cp r17,r16
	ldd r16,Y+last_sect-fs+1
	cpc r18,r16
	ldd r16,Y+last_sect-fs+2
	cpc r19,r16
#ifdef _FS_FAT32
	lds r16,(last_sect+3)
#else
	ldd r16,Y+last_sect-fs+3
#endif
	cpc r20,r16
	brne disk_read02
	rjmp disk_read_x
disk_read02:
	std Y+last_sect-fs,r17
	std Y+last_sect-fs+1,r18
	std Y+last_sect-fs+2,r19
#ifdef _FS_FAT32
	sts (last_sect+3),r20
#else
	std Y+last_sect-fs+3,r20
#endif
	sbi cache_flag
disk_read01:
	ldi r16,1
disk_read0:
#endif
#ifdef cache_sector
	cbi valid_sect
#endif
	tst r16
	brne disk_read1
	ldi r16,RES_PARERR
	rjmp disk_read_x
disk_read1:
	push r16
	lds r16,(cardtype)
	andi r16,CT_BLOCK
	brne disk_read2
	mov r20,r19
	mov r19,r18
	mov r18,r17
	clr r17
	add r18,r18
	adc r19,r19
	adc r20,r20
disk_read2:
	push r17
	push r18
	push r19
	push r20
	rcall disk_status
	pop r20
	pop r19
	pop r18
	pop r17
	pop r21
	andi r16,STA_NOINIT
	breq disk_read3
	ldi r16,RES_NOTRDY
	rjmp disk_read_x
disk_read3:
	ldi r16,CMD17
	cpi r21,1
	breq disk_read4
	ldi r16,CMD18
disk_read4:
	push r16
	rcall send_cmd
	pop r17
	tst r16
	brne disk_read_x1
#ifdef cache_sector
	push r21
disk_read_lp:
	ldi r18,LOW(512)
	ldi r19,HIGH(512)
	rcall rcvr_datablock
	tst r16
	breq disk_read_x2
	dec r21
	brne disk_read_lp
disk_read_x2:
	cpi r17,CMD18
	brne disk_read_x3
	ldi r16,CMD12
	rcall send_cmd_p0
disk_read_x3:
#ifdef cache_sector
	pop r16
	cpi r16,1
	brne disk_read_x1
	sbic cache_flag
	sbi valid_sect
#endif
disk_read_x1:
	call deselect
	ldi r16,RES_OK
	tst r21
	breq disk_read_x
	ldi r16,RES_ERROR
disk_read_x:
	ret
#endif


;table needed for prescaling according to Arai's IDCT algorithm in zigzag order
arai_scale_zz:
.dw int(1.38704*32768) ;1
.dw int(1.38704*32768) ;2
.dw int(1.30656*32768) ;3
.dw int(1.92388*32768) ;4
.dw int(1.30656*32768) ;5
.dw int(1.17588*32768) ;6
.dw int(1.81226*32768) ;7
.dw int(1.81226*32768) ;8
.dw int(1.17588*32768) ;9
.dw int(1.00000*32768) ;10
.dw int(1.63099*32768) ;11
.dw int(1.70711*32768) ;12
.dw int(1.63099*32768) ;13
.dw int(1.00000*32768) ;14
.dw int(0.78570*32768) ;15
.dw int(1.38704*32768) ;16
.dw int(1.53636*32768) ;17
.dw int(1.53636*32768) ;18
.dw int(1.38704*32768) ;19
.dw int(0.78570*32768) ;20
.dw int(0.54120*32768) ;21
.dw int(1.08979*32768) ;22
.dw int(1.30656*32768) ;23
.dw int(1.38268*32768) ;24
.dw int(1.30656*32768) ;25
.dw int(1.08979*32768) ;26
.dw int(0.54120*32768) ;27
.dw int(0.27590*32768) ;28
.dw int(0.75066*32768) ;29
.dw int(1.02656*32768) ;30
.dw int(1.17588*32768) ;31
.dw int(1.17588*32768) ;32
.dw int(1.02656*32768) ;33
.dw int(0.75066*32768) ;34
.dw int(0.27590*32768) ;35
.dw int(0.38268*32768) ;36
.dw int(0.70711*32768) ;37
.dw int(0.92388*32768) ;38
.dw int(1.00000*32768) ;39
.dw int(0.92388*32768) ;40
.dw int(0.70711*32768) ;41
.dw int(0.38268*32768) ;42
.dw int(0.36048*32768) ;43
.dw int(0.63638*32768) ;44
.dw int(0.78570*32768) ;45
.dw int(0.78570*32768) ;46
.dw int(0.63638*32768) ;47
.dw int(0.36048*32768) ;48
.dw int(0.32442*32768) ;49
.dw int(0.54120*32768) ;50
.dw int(0.61732*32768) ;51
.dw int(0.54120*32768) ;52
.dw int(0.32442*32768) ;53
.dw int(0.27590*32768) ;54
.dw int(0.42522*32768) ;55
.dw int(0.42522*32768) ;56
.dw int(0.27590*32768) ;57
.dw int(0.21678*32768) ;58
.dw int(0.29290*32768) ;59
.dw int(0.21677*32768) ;60
.dw int(0.14932*32768) ;61
.dw int(0.14932*32768) ;62
.dw int(0.07612*32768) ;63

;table with power of 2 values
pow2tab: .dw 1,2,4,8,16,32,64,128,256,512,1024

ep_power_off_str:
	.db 2,VCOM_AND_DATA_INTERVAL_SETTING,$f7,1,POWER_OFF,1,WT_IDLE,0
ep_sleep_str:
	.db 2,DEEP_SLEEP,$a5,0
ep_init_std:
	.db 4,BOOSTER_SOFT_START,$17,$17,$17,1,POWER_ON,1,WT_IDLE,2,PANEL_SETTING,$0f,0  ,-1	;LUT from OTP
ep_init_volt_str:
	.db 2,PANEL_SETTING,$3f,2,PLL_CONTROL,$3A,6,POWER_SETTING,3,0,$2b,$2b,0,2,VCM_DC_SETTING,25,2,TCON_SETTING,$62,2,VCOM_AND_DATA_INTERVAL_SETTING,0b11010111,0   ,-1

.equ grays1x=3
.equ pauses1x=1
ep_lut_grayxs_str:
	.db 45,LUT_FOR_VCOM
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db $00,$00
	.db 43,LUT_WHITE_TO_WHITE
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_WHITE
	.db $40,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_WHITE_TO_BLACK
	.db $40,grays1x,pauses1x,0,0,1
	.db $40,grays1x,pauses1x,0,0,1
	.db 0,grays1x,pauses1x,0,0,1
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_BLACK
	.db $40,grays1x,pauses1x,0,0,1
	.db $40,grays1x,pauses1x,0,0,1
	.db $40,grays1x,pauses1x,0,0,1
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0  ,-1

.equ grays2x=3
.equ pauses2x=1
ep_lut_grayxs1_str:
	.db 45,LUT_FOR_VCOM
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,0,0,0,0,0
	.db $00,$00
	.db 43,LUT_WHITE_TO_WHITE
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_WHITE
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,0,0,0,0,0
	.db 43,LUT_WHITE_TO_BLACK
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_BLACK
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db $44,grays2x,pauses2x,grays2x,pauses2x,1
	.db 0,0,0,0,0,0
	.db 0  ,-1

ep_lut_clr:
	.db 4,BOOSTER_SOFT_START,$17,$17,$17,1,POWER_ON,1,WT_IDLE,2,PANEL_SETTING,$2f
	.db 2,PLL_CONTROL,$3A,6,POWER_SETTING,3,0,$2b,$2b,4,3,VCM_DC_SETTING,$16,$16
	.db 45,LUT_FOR_VCOM
	.db $00,$4b,$4b,$00,$00,1
	.db $00,$0D,$0D,$0D,$0D,1
	.db $00,$0A,$0A,$05,$05,3
	.db $00,$05,$05,$03,$03,2
	.db $00,$0D,$0A,$0D,$0A,3
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 0,0
	.db 43,LUT_WHITE_TO_WHITE
	.db $90,$4b,$4b,$00,$00,1
	.db $44,$0D,$0D,$0D,$0D,1
	.db $66,$0A,$0A,$05,$05,3
	.db $99,$05,$05,$03,$03,2
	.db $88,$0D,$0A,$0D,$0A,3
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_WHITE
	.db $90,$4b,$4b,$00,$00,1
	.db $44,$0D,$0D,$0D,$0D,1
	.db $66,$0A,$0A,$05,$05,3
	.db $99,$05,$05,$03,$03,2
	.db $88,$0D,$0A,$0D,$0A,3
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_WHITE_TO_BLACK
	.db $90,$4b,$4b,$00,$00,1
	.db $44,$0D,$0D,$0D,$0D,1
	.db $66,$0A,$0A,$05,$05,3
	.db $99,$05,$05,$03,$03,2
	.db $88,$0D,$0A,$0D,$0A,3
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 43,LUT_BLACK_TO_BLACK
	.db $90,$4b,$4b,$00,$00,1
	.db $44,$0D,$0D,$0D,$0D,1
	.db $66,$0A,$0A,$05,$05,3
	.db $99,$05,$05,$03,$03,2
	.db $88,$0D,$0A,$0D,$0A,3
	.db 0,0,0,0,0,0
	.db 0,0,0,0,0,0
	.db 1,DISPLAY_REFRESH,0  ,-1 

insert_card_pic:
	.include "InsertCard1.inc"

lowbatt_pic:
	.include "LowBatt.inc"

	.dseg
	.org SRAM_START
;memory storage needed for decoding
;the arrangement of the fields must be preserved to make sure address arithmetics will work properly
sect_buf: .byte rd_buf_size		;must start at 256 byte boundary
blk_y: .byte 128
tmp_buf: .byte 128
q_tbl:	.byte 64
prv_dc: .byte 2
ht_ptr: .byte 2*2
x_ptr: .byte 1
y_ptr: .byte 1
wd_cnt: .byte 1
rd_ptr: .byte 2
w_time: .byte 1
random: .byte 4
lowbatt_cnt: .byte 1

dstat: .byte 1
cardtype: .byte 1
sector:	.byte 512
fs:
fs_type: .byte 1		/* FAT sub type */
flag:	 .byte 1		/* File status flags */
csize:	 .byte 1		/* Number of sectors per cluster */
n_rootdir: .byte 2		/* Number of root directory entries (0 on FAT32) */
n_fatent:.byte clust_sz	/* Number of FAT entries (= number of clusters + 2) */
fatbase: .byte 4		/* FAT start sector */
fatsize: .byte 4		/* Number of sectors per FAT */
dirbase: .byte 4		/* Root directory start sector (Cluster# on FAT32) */
database:.byte 4		/* Data start sector */
fptr:	 .byte 4		/* File R/W pointer */
fsize:	 .byte 4		/* File size */
org_clust: .byte clust_sz	/* File start cluster */
curr_clust:.byte clust_sz	/* File current cluster */
dsect:	.byte 4			/* File current data sector */
dj:
index:	.byte 2			/* Current read/write index number */
fn:		.byte 2			/* Pointer to the SFN (in/out) {file[8],ext[3],status[1]} */
sclust:	.byte clust_sz	/* Table start cluster (0:Static table) */
clust:	.byte clust_sz	/* Current cluster */
sect:	.byte 4			/* Current sector */
#ifdef cache_sector
last_sect: .byte 4
#endif
pict_file_ptr: .byte 4
buffer:
ht:	.byte ht_size
recent_tab: .byte recent_entries*4

